from typing import TYPE_CHECKING, Any, Union, cast

from . import gitlab

if TYPE_CHECKING:
    from . import merge_request as mb_merge_request


class Approvals(gitlab.Resource):
    """Approval info for a MergeRequest.
    https://docs.gitlab.com/ee/api/merge_request_approvals.html"""

    def refetch_info(self) -> None:
        approver_url = (
            f"/projects/{self.project_id}/merge_requests/{self.iid}/approvals"
        )
        info = self._api.call(gitlab.GET(approver_url))

        if TYPE_CHECKING:
            assert isinstance(info, dict)
        self._info = info

    @property
    def id(self) -> int:
        raise NotImplementedError()

    @property
    def iid(self) -> int:
        return cast(int, self.info["iid"])

    @property
    def project_id(self) -> int:
        return cast(int, self.info["project_id"])

    @property
    def approvals_left(self) -> int:
        return cast(int, self.info.get("approvals_left", 0))

    @property
    def sufficient(self) -> bool:
        return not self.approvals_left

    @property
    def approver_usernames(self) -> list[str]:
        return [who["user"]["username"] for who in self.info["approved_by"]]

    @property
    def approver_ids(self) -> list[int]:
        """Return the uids of the approvers."""
        return [who["user"]["id"] for who in self.info["approved_by"]]

    def reapprove(self) -> None:
        """Impersonates the approvers and re-approves the merge_request as them.

        The idea is that we want to get the approvers, push the rebased branch
        (which may invalidate approvals, depending on GitLab settings) and then
        restore the approval status.
        """
        self.approve(self)

    def approve(self, obj: Union["Approvals", "mb_merge_request.MergeRequest"]) -> None:
        """Approve an object which can be a merge_request or an approval."""
        approve_url = f"/projects/{obj.project_id}/merge_requests/{obj.iid}/approve"

        for uid in self.approver_ids:
            self._api.call(gitlab.POST(approve_url), sudo=uid)


class CustomApprovals(Approvals):
    """Allows a limited way to manage approvals on CE."""

    def __init__(
        self,
        api: gitlab.Api,
        info: dict[str, Any],
        allowed_approvers: list[str],
        approvals_required: int = 1,
    ):
        super().__init__(api, info)
        self._allowed_approvers = allowed_approvers
        self._approvals_required = approvals_required

    @property
    def allowed_approvals_usernames(self) -> list[str]:
        """Filter approver_usernames to only those allowed."""
        return [
            username
            for username in self.approver_usernames
            if username in self._allowed_approvers
        ]

    @property
    def approvals_left(self) -> int:
        return max(0, self._approvals_required - len(self.allowed_approvals_usernames))

    @property
    def sufficient(self) -> bool:
        return not self.approvals_left



from typing import TYPE_CHECKING, Any, Optional, cast

from . import gitlab

GET, POST = gitlab.GET, gitlab.POST


class Pipeline(gitlab.Resource):
    def __init__(self, api: gitlab.Api, info: dict[str, Any], project_id: int):
        info["project_id"] = project_id
        super().__init__(api, info)

    @classmethod
    def pipelines_by_branch(
        cls,
        project_id: int,
        branch: str,
        api: gitlab.Api,
        *,
        ref: Optional[str] = None,
        status: Optional[str] = None,
        order_by: str = "id",
        sort: str = "desc",
    ) -> list["Pipeline"]:
        params = {
            "ref": branch if ref is None else ref,
            "order_by": order_by,
            "sort": sort,
        }
        if status is not None:
            params["status"] = status
        pipelines_info = api.call(GET(f"/projects/{project_id}/pipelines", params))
        if TYPE_CHECKING:
            assert isinstance(pipelines_info, list)

        return [cls(api, pipeline_info, project_id) for pipeline_info in pipelines_info]

    @classmethod
    def pipelines_by_merge_request(
        cls, project_id: int, merge_request_iid: int, api: gitlab.Api
    ) -> list["Pipeline"]:
        """Fetch all pipelines for a merge request in descending order of
        pipeline ID."""
        pipelines_info = api.call(
            GET(f"/projects/{project_id}/merge_requests/{merge_request_iid}/pipelines")
        )
        if TYPE_CHECKING:
            assert isinstance(pipelines_info, list)
        pipelines_info.sort(
            key=lambda pipeline_info: cast(str, pipeline_info["id"]), reverse=True
        )
        return [cls(api, pipeline_info, project_id) for pipeline_info in pipelines_info]

    @classmethod
    def manual_jobs_by_pipeline(
        cls, project_id: int, pipeline_id: int, api: Any
    ) -> Any:
        jobs = api.collect_all_pages(
            GET(f"/projects/{project_id}/pipelines/{pipeline_id}/jobs?scope[]=manual")
        )
        return jobs

    @property
    def project_id(self) -> int:
        result = self.info["project_id"]
        if TYPE_CHECKING:
            assert isinstance(result, int)
        return result

    @property
    def id(self) -> int:
        result = self.info["id"]
        if TYPE_CHECKING:
            assert isinstance(result, int)
        return result

    @property
    def status(self) -> str:
        result = self.info["status"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def ref(self) -> str:
        result = self.info["ref"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def sha(self) -> str:
        result = self.info["sha"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def web_url(self) -> str:
        result = self.info["web_url"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    def cancel(self) -> dict[str, Any]:
        result = self._api.call(
            POST(f"/projects/{self.project_id}/pipelines/{self.id}/cancel")
        )
        if TYPE_CHECKING:
            assert isinstance(result, dict)
        return result



from unittest.mock import Mock

from marge.gitlab import GET, Api
from marge.user import User

INFO = {
    "id": 1234,
    "username": "john_smith",
    "name": "John Smith",
    "state": "active",
    "is_admin": False,
}


# pylint: disable=attribute-defined-outside-init
class TestProjectWithUser:
    def setup_method(self, _method):
        self.api = Mock(Api)

    def test_fetch_myself(self):
        User.myself(api=self.api)
        self.api.call.assert_called_once_with(GET("/user"))

    def test_fetch_by_id(self):
        api = self.api
        api.call = Mock(return_value=INFO)

        user = User.fetch_by_id(user_id=1234, api=api)

        api.call.assert_called_once_with(GET("/users/1234"))
        assert user.info == INFO

    def test_properties(self):
        user = User(api=self.api, info=INFO)
        assert user.id == 1234
        assert user.username == "john_smith"
        assert user.name == "John Smith"
        assert user.state == "active"



import datetime
import enum
import operator
import re
import zoneinfo
from collections.abc import Iterable
from typing import Union


# pylint: disable=invalid-name
@enum.unique
class WeekDay(enum.Enum):
    Monday = 0
    Tuesday = 1
    Wednesday = 2
    Thursday = 3
    Friday = 4
    Saturday = 5
    Sunday = 6


_DAY_NAMES = {day.name.lower(): day for day in WeekDay}
_DAY_NAMES.update((day.name.lower()[:3], day) for day in WeekDay)


def find_weekday(string_or_day: Union[WeekDay, str]) -> WeekDay:
    if isinstance(string_or_day, WeekDay):
        return string_or_day

    if isinstance(string_or_day, str):
        return _DAY_NAMES[string_or_day.lower()]

    raise ValueError(f"Not a week day: {string_or_day!r}")


def parse_time(time_str: str, tz_str: str) -> datetime.datetime:
    time_re = re.compile(
        # Hour as either 1 or 2 digits
        r"(?P<hour>\d{1,2})"
        # Optional minute as :MM
        r"(?::(?P<minute>\d{2}))?"
        # Optional AM/PM
        "(?P<ampm>[AaPp][Mm])?"
    )
    m = time_re.fullmatch(time_str)
    if not m:
        raise ValueError(f"Could not parse time string: {time_str}")

    hour = int(m.group("hour"))
    minute = int(m.group("minute") or 0)
    ampm = m.group("ampm")

    if ampm:
        ampm = ampm.lower()
        if ampm == "am" and hour == 12:
            hour = 0
        elif ampm == "pm" and hour < 12:
            hour += 12

    dt = datetime.datetime.now(zoneinfo.ZoneInfo(tz_str))
    return dt.replace(hour=hour, minute=minute, second=0, microsecond=0)


class WeeklyInterval:
    def __init__(
        self,
        from_weekday: Union[WeekDay, str],
        from_time: datetime.time,
        to_weekday: Union[WeekDay, str],
        to_time: datetime.time,
    ):
        from_weekday = find_weekday(from_weekday)
        to_weekday = find_weekday(to_weekday)

        # the class invariant is that from_weekday <= to_weekday; so when this
        # is not the case (e.g. a Fri-Mon interval), we store the complement interval
        # (in the example, Mon-Fri), and invert the criterion
        self._is_complement_interval = from_weekday.value > to_weekday.value
        if self._is_complement_interval:
            self._from_weekday = to_weekday
            self._from_time = to_time
            self._to_weekday = from_weekday
            self._to_time = from_time
        else:
            self._from_weekday = from_weekday
            self._from_time = from_time
            self._to_weekday = to_weekday
            self._to_time = to_time

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, WeeklyInterval):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: object) -> bool:
        return not self == other

    def __repr__(self) -> str:
        pat = "{class_name}({from_weekday}, {from_time}, {to_weekday}, {to_time})"
        if self._is_complement_interval:
            return pat.format(
                class_name=self.__class__.__name__,
                from_weekday=self._to_weekday,
                from_time=self._to_time,
                to_weekday=self._from_weekday,
                to_time=self._from_time,
            )
        return pat.format(
            class_name=self.__class__.__name__,
            from_weekday=self._from_weekday,
            from_time=self._from_time,
            to_weekday=self._to_weekday,
            to_time=self._to_time,
        )

    @classmethod
    def from_human(cls, string: str) -> "WeeklyInterval":
        from_, to_ = string.split("-")

        def parse_part(part: str) -> tuple[WeekDay, datetime.time]:
            part = part.replace("@", " ")
            parts = part.split()
            day_part = parts[0]
            time_part = parts[1]
            zone_part = parts[2] if len(parts) > 2 else "UTC"
            weekday = find_weekday(day_part)
            dt = parse_time(time_part, zone_part)
            time = dt.astimezone(datetime.timezone.utc).time()
            return weekday, time

        from_weekday, from_time = parse_part(from_)
        to_weekday, to_time = parse_part(to_)
        return cls(from_weekday, from_time, to_weekday, to_time)

    def covers(self, date: datetime.datetime) -> bool:
        return self._interval_covers(date) != self._is_complement_interval

    def _interval_covers(self, date: datetime.datetime) -> bool:
        weekday = date.date().weekday()
        time = date.time()
        before = operator.le if self._is_complement_interval else operator.lt

        if not self._from_weekday.value <= weekday <= self._to_weekday.value:
            return False

        if self._from_weekday.value == weekday and before(time, self._from_time):
            return False

        if self._to_weekday.value == weekday and before(self._to_time, time):
            return False

        return True


class IntervalUnion:
    def __init__(self, iterable: Iterable[WeeklyInterval]):
        self._intervals = list(iterable)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, IntervalUnion):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: object) -> bool:
        return not self == other

    def __repr__(self) -> str:
        return "{o.__class__.__name__}({o._intervals})".format(o=self)

    @classmethod
    def empty(cls) -> "IntervalUnion":
        return cls(())

    @classmethod
    def from_human(cls, string: str) -> "IntervalUnion":
        strings = string.split(",")
        return cls(WeeklyInterval.from_human(s) for s in strings)

    def covers(self, date: datetime.datetime) -> bool:
        return any(interval.covers(date) for interval in self._intervals)



import logging as log
import time
from typing import Optional

from . import branch, git, gitlab, job
from . import project as mb_project
from . import user as mb_user
from .commit import Commit
from .merge_request import MergeRequest
from .pipeline import Pipeline


class CannotBatch(Exception):
    pass


class BatchMergeJob(job.MergeJob):
    def __init__(
        self,
        *,
        api: gitlab.Api,
        user: mb_user.User,
        project: mb_project.Project,
        repo: git.Repo,
        options: job.MergeJobOptions,
        merge_requests: list[MergeRequest],
        batch_branch_name: str,
    ):
        super().__init__(
            api=api, user=user, project=project, repo=repo, options=options
        )
        self.batch_branch_name = batch_branch_name
        self._merge_requests = merge_requests

    def remove_local_batch_branch(self) -> None:
        log.info("Removing local batch branch")
        try:
            self._repo.remove_branch(self.batch_branch_name)
        except git.GitError:
            pass

    def remove_remote_batch_branch(self) -> None:
        log.info("Removing remote batch branch")
        try:
            branch.Branch.delete_by_name(
                project_id=self._project.id,
                branch=self.batch_branch_name,
                api=self._api,
            )
        except gitlab.NotFound:
            pass
        except gitlab.ApiError as err:
            log.warning("Failed to remove remote batch branch: %s", err)

    def close_batch_mr(self) -> None:
        log.info("Closing batch MRs")
        params = {
            "author_id": self._user.id,
            "labels": self.batch_branch_name,
            "state": "opened",
            "order_by": "created_at",
            "sort": "desc",
        }
        batch_mrs = MergeRequest.search(
            api=self._api, project_id=self._project.id, params=params
        )
        for batch_mr in batch_mrs:
            log.info("Closing batch MR !%s", batch_mr.iid)
            batch_mr.close()
            for batch_pipeline in Pipeline.pipelines_by_merge_request(
                self._project.id, batch_mr.iid, self._api
            ):
                if batch_pipeline.status in ("pending", "running"):
                    log.info("Cancelling obsolete batch pipeline %s", batch_pipeline.id)
                    batch_pipeline.cancel()

    def create_batch_mr(self, target_branch: str) -> MergeRequest:
        self.push_batch()
        log.info("Creating batch MR")
        params = {
            "source_branch": self.batch_branch_name,
            "target_branch": target_branch,
            "title": "Marge Bot Batch MR - DO NOT TOUCH",
            "labels": self.batch_branch_name,
        }
        batch_mr = MergeRequest.create(
            api=self._api, project_id=self._project.id, params=params
        )
        log.info("Batch MR !%s created", batch_mr.iid)
        return batch_mr

    def get_mrs_with_common_target_branch(
        self, target_branch: str
    ) -> list[MergeRequest]:
        log.info("Filtering MRs with target branch %s", target_branch)
        return [
            merge_request
            for merge_request in self._merge_requests
            if merge_request.target_branch == target_branch
        ]

    def ensure_mergeable_mr(
        self, merge_request: MergeRequest, skip_ci: bool = False
    ) -> None:
        super().ensure_mergeable_mr(merge_request)

        if self._project.only_allow_merge_if_pipeline_succeeds and not skip_ci:
            ci_status, pipeline_msg, _ = self.get_mr_ci_status(merge_request)
            if ci_status != "success":
                raise CannotBatch(f"This MR has not passed CI. {pipeline_msg}")

    def get_mergeable_mrs(
        self, merge_requests: list[MergeRequest]
    ) -> list[MergeRequest]:
        log.info("Filtering mergeable MRs")
        mergeable_mrs = []
        for merge_request in merge_requests:
            try:
                self.ensure_mergeable_mr(merge_request)
            except (CannotBatch, job.SkipMerge) as ex:
                log.warning('Skipping unbatchable MR: "%s"', ex)
            except job.CannotMerge as ex:
                log.warning('Skipping unmergeable MR: "%s"', ex)
                self.unassign_from_mr(merge_request)
                merge_request.comment(f"I couldn't merge this branch: {ex}")
            else:
                mergeable_mrs.append(merge_request)
        return mergeable_mrs

    def push_batch(self) -> None:
        log.info("Pushing batch branch")
        self._repo.push(self.batch_branch_name, force=True)

    def ensure_mr_not_changed(self, merge_request: MergeRequest) -> None:
        log.info("Ensuring MR !%s did not change", merge_request.iid)
        changed_mr = MergeRequest.fetch_by_iid(
            merge_request.project_id, merge_request.iid, self._api
        )
        error_message = "The {} changed whilst merging!"
        for attr in (
            "source_branch",
            "source_project_id",
            "target_branch",
            "target_project_id",
            "sha",
        ):
            if getattr(changed_mr, attr) != getattr(merge_request, attr):
                raise job.CannotMerge(error_message.format(attr.replace("_", " ")))

    def merge_batch(
        self, target_branch: str, source_branch: str, no_ff: bool = False
    ) -> str:
        if no_ff:
            return self._repo.merge(target_branch, source_branch, "--no-ff")

        return self._repo.fast_forward(target_branch, source_branch)

    def update_merge_request(
        self, merge_request: MergeRequest, source_repo_url: Optional[str] = None
    ) -> str:
        log.info("Fusing MR !%s", merge_request.iid)
        approvals = merge_request.fetch_approvals(self.approvals_factory)

        _, _, actual_sha = self.update_from_target_branch_and_push(
            merge_request,
            source_repo_url=source_repo_url,
            skip_ci=self._options.skip_ci_batches,
        )

        sha_now = Commit.last_on_branch(
            merge_request.source_project_id, merge_request.source_branch, self._api
        ).id
        log.info(
            "update_merge_request: sha_now (%s), actual_sha (%s)", sha_now, actual_sha
        )
        # Make sure no-one managed to race and push to the branch in the
        # meantime, because we're about to impersonate the approvers, and
        # we don't want to approve unreviewed commits
        if sha_now != actual_sha:
            raise job.CannotMerge(
                "Someone pushed to branch while we were trying to merge"
            )

        # As we're not using the API to merge the individual MR, we don't strictly need to reapprove it.
        # However, it's a little weird to look at the merged MR to find it has no approvals,
        # so let's do it anyway.
        self.maybe_reapprove(merge_request, approvals)
        return sha_now

    def accept_mr(
        self,
        merge_request: MergeRequest,
        expected_remote_target_branch_sha: str,
        source_repo_url: Optional[str] = None,
    ) -> str:
        log.info("Accept MR !%s", merge_request.iid)

        # Make sure latest commit in remote <target_branch> is the one we tested against
        new_target_sha = Commit.last_on_branch(
            self._project.id, merge_request.target_branch, self._api
        ).id
        if new_target_sha != expected_remote_target_branch_sha:
            raise CannotBatch("Someone was naughty and by-passed marge")

        # Rebase and apply the trailers
        self.update_merge_request(merge_request, source_repo_url=source_repo_url)

        # This switches git to <target_branch>
        final_sha = self.merge_batch(
            merge_request.target_branch,
            merge_request.source_branch,
            self._options.use_no_ff_batches,
        )
        # Don't force push in case the remote has changed.
        self._repo.push(merge_request.target_branch, force=False)

        time.sleep(2)

        # At this point Gitlab should have recognised the MR as being accepted.
        log.info("Successfully merged MR !%s", merge_request.iid)

        pipelines = Pipeline.pipelines_by_branch(
            api=self._api,
            project_id=merge_request.source_project_id,
            branch=merge_request.source_branch,
            status="running",
        )
        for pipeline in pipelines:
            pipeline.cancel()

        if merge_request.force_remove_source_branch:
            branch.Branch.delete_by_name(
                project_id=merge_request.source_project_id,
                branch=merge_request.source_branch,
                api=self._api,
            )

        return final_sha

    def execute(self, exc_comment: Optional[str] = None) -> None:
        # pylint: disable=too-many-branches,too-many-statements
        # Cleanup previous batch work
        self.remove_local_batch_branch()
        self.close_batch_mr()

        target_branch = self._merge_requests[0].target_branch
        merge_requests = self.get_mrs_with_common_target_branch(target_branch)
        merge_requests = self.get_mergeable_mrs(merge_requests)

        if len(merge_requests) <= 1:
            # Either no merge requests are ready to be merged, or there's only one for this target branch.
            # Let's raise an error to do a basic job for these cases.
            raise CannotBatch("not enough ready merge requests")

        self._repo.fetch("origin")

        # Save the sha of remote <target_branch> so we can use it to make sure
        # the remote wasn't changed while we're testing against it
        remote_target_branch_sha = self._repo.get_commit_hash(f"origin/{target_branch}")

        self._repo.checkout_branch(target_branch, f"origin/{target_branch}")
        self._repo.checkout_branch(self.batch_branch_name, f"origin/{target_branch}")

        batch_mr = self.create_batch_mr(target_branch=target_branch)
        batch_mr_sha = batch_mr.sha

        working_merge_requests = []

        for merge_request in merge_requests:
            try:
                _, source_repo_url, merge_request_remote = self.fetch_source_project(
                    merge_request
                )
                self._repo.checkout_branch(
                    merge_request.source_branch,
                    f"{merge_request_remote}/{merge_request.source_branch}",
                )

                if self._options.use_merge_commit_batches:
                    # Rebase and apply the trailers before running the batch MR
                    actual_sha = self.update_merge_request(
                        merge_request, source_repo_url=source_repo_url
                    )
                    # Update <batch> branch with MR changes
                    batch_mr_sha = self._repo.merge(
                        self.batch_branch_name,
                        merge_request.source_branch,
                        "-m",
                        f"Batch merge !{merge_request.iid} into"
                        + f"{merge_request.target_branch} (!{batch_mr.iid})",
                        local=True,
                    )
                else:
                    # Update <source_branch> on latest <batch> branch so it contains previous MRs
                    self.fuse(
                        merge_request.source_branch,
                        self.batch_branch_name,
                        source_repo_url=source_repo_url,
                        local=True,
                    )
                    # Update <batch> branch with MR changes
                    batch_mr_sha = self._repo.fast_forward(
                        self.batch_branch_name, merge_request.source_branch, local=True
                    )

                # We don't need <source_branch> anymore. Remove it now in case another
                # merge request is using the same branch name in a different project.
                self._repo.remove_branch(merge_request.source_branch)
            except (git.GitError, job.CannotMerge):
                log.warning(
                    "Skipping MR !%s, got conflicts while rebasing", merge_request.iid
                )
                continue
            else:
                if self._options.use_merge_commit_batches:
                    # update merge_request with the current sha, we will compare it with
                    # the actual sha later to make sure no one pushed this MR meanwhile
                    merge_request.update_sha(actual_sha)

                working_merge_requests.append(merge_request)

        if len(working_merge_requests) <= 1:
            raise CannotBatch("not enough ready merge requests")

        # This switches git to <batch> branch
        self.push_batch()
        for merge_request in working_merge_requests:
            merge_request.comment(
                f"I will attempt to batch this MR (!{batch_mr.iid})..."
            )

        # wait for the CI of the batch MR
        if self._project.only_allow_merge_if_pipeline_succeeds:
            try:
                self.wait_for_ci_to_pass(batch_mr, commit_sha=batch_mr_sha)
            except job.CannotMerge as err:
                for merge_request in working_merge_requests:
                    merge_request.comment(
                        f"Batch MR !{batch_mr.iid} failed: {err.reason} I will retry later..."
                    )
                raise CannotBatch(err.reason) from err

        # check each sub MR, and accept each sub MR if using the normal batch
        for merge_request in working_merge_requests:
            try:
                # FIXME: this should probably be part of the merge request
                _, source_repo_url, merge_request_remote = self.fetch_source_project(
                    merge_request
                )
                self.ensure_mr_not_changed(merge_request)
                # we know the batch MR's CI passed, so we skip CI for sub MRs this time
                self.ensure_mergeable_mr(merge_request, skip_ci=True)

                if not self._options.use_merge_commit_batches:
                    # accept each MRs
                    remote_target_branch_sha = self.accept_mr(
                        merge_request,
                        remote_target_branch_sha,
                        source_repo_url=source_repo_url,
                    )
            except CannotBatch as err:
                merge_request.comment(
                    f"I couldn't merge this branch: {str(err)} I will retry later..."
                )
                raise
            except job.SkipMerge:
                # Raise here to avoid being caught below - we don't want to be unassigned.
                raise
            except job.CannotMerge as err:
                self.unassign_from_mr(merge_request)
                merge_request.comment(f"I couldn't merge this branch: {err.reason}")
                raise

        # Accept the batch MR
        if self._options.use_merge_commit_batches:
            # Approve the batch MR using the last sub MR's approvers
            if not batch_mr.fetch_approvals(self.approvals_factory).sufficient:
                approvals = working_merge_requests[-1].fetch_approvals(
                    self.approvals_factory
                )
                try:
                    approvals.approve(batch_mr)
                except (gitlab.Forbidden, gitlab.Unauthorized):
                    log.exception("Failed to approve MR:")

            try:
                ret = batch_mr.accept(
                    remove_branch=batch_mr.force_remove_source_branch,
                    sha=batch_mr_sha,
                    merge_when_pipeline_succeeds=bool(
                        self._project.only_allow_merge_if_pipeline_succeeds
                    ),
                )
                log.debug("batch_mr.accept result: %s", ret)
            except gitlab.ApiError as err:
                log.exception("Gitlab API Error:")
                raise job.CannotMerge(f"Gitlab API Error: {err}") from err

        # Cleanup current batch work
        self.remove_local_batch_branch()
        self.remove_remote_batch_branch()



from unittest.mock import Mock, call, patch

import pytest

import marge.user
from marge.approvals import Approvals, CustomApprovals
from marge.gitlab import GET, POST, Api

# testing this here is more convenient
from marge.job import CannotMerge, _get_reviewer_names_and_emails
from marge.merge_request import MergeRequest

INFO = {
    "id": 5,
    "iid": 6,
    "project_id": 1,
    "title": "Approvals API",
    "description": "Test",
    "state": "opened",
    "created_at": "2016-06-08T00:19:52.638Z",
    "updated_at": "2016-06-08T21:20:42.470Z",
    "detailed_merge_status": "mergeable",
    "approvals_required": 3,
    "approvals_left": 1,
    "approved_by": [
        {
            "user": {
                "name": "Administrator",
                "username": "root",
                "id": 1,
                "state": "active",
                "avatar_url": "".join(
                    [
                        "http://www.gravatar.com/avatar/",
                        "e64c7d89f26bd1972efa854d13d7dd61?s=80\u0026d=identicon",
                    ]
                ),
                "web_url": "http://localhost:3000/u/root",
            }
        },
        {
            "user": {
                "name": "Roger Ebert",
                "username": "ebert",
                "id": 2,
                "state": "active",
            }
        },
    ],
}
USERS = {
    1: {
        "name": "Administrator",
        "username": "root",
        "id": 1,
        "state": "active",
        "email": "root@localhost",
    },
    2: {
        "name": "Roger Ebert",
        "username": "ebert",
        "id": 2,
        "state": "active",
        "email": "ebert@example.com",
    },
}


# pylint: disable=attribute-defined-outside-init
class TestApprovals:
    def setup_method(self, _method):
        self.api = Mock(Api)
        self.approvals = Approvals(api=self.api, info=INFO)

    def test_fetch_from_merge_request(self):
        api = self.api
        api.call = Mock(return_value=INFO)

        merge_request = MergeRequest(api, {"id": 74, "iid": 6, "project_id": 1234})
        approvals = merge_request.fetch_approvals(Approvals)

        api.call.assert_called_once_with(
            GET("/projects/1234/merge_requests/6/approvals")
        )
        assert approvals.info == INFO

    def test_properties(self):
        assert self.approvals.project_id == 1
        assert self.approvals.approvals_left == 1
        assert self.approvals.approver_usernames == ["root", "ebert"]
        assert not self.approvals.sufficient

    def test_sufficiency(self):
        good_approvals = Approvals(
            api=self.api, info=dict(INFO, approvals_required=1, approvals_left=0)
        )
        assert good_approvals.sufficient

    def test_custom_insufficient(self):
        api = self.api
        api.call = Mock(return_value=INFO)

        merge_request = MergeRequest(api, {"id": 74, "iid": 6, "project_id": 1234})

        def make_custom_approvals(api, info):
            # Only require 2 approvers from the list, not actually achievable
            return CustomApprovals(api, info, ["root"], 2)

        approvals = merge_request.fetch_approvals(make_custom_approvals)

        api.call.assert_called_once_with(
            GET("/projects/1234/merge_requests/6/approvals")
        )

        assert approvals.info == INFO
        assert approvals.approvals_left == 1
        assert not approvals.sufficient

    def test_custom_sufficient(self):
        api = self.api
        api.call = Mock(return_value=INFO)

        merge_request = MergeRequest(api, {"id": 74, "iid": 6, "project_id": 1234})

        def make_custom_approvals(api, info):
            # Only require 1 approver from the list
            return CustomApprovals(api, info, ["root"], 1)

        approvals = merge_request.fetch_approvals(make_custom_approvals)

        api.call.assert_called_once_with(
            GET("/projects/1234/merge_requests/6/approvals")
        )

        assert approvals.info == INFO
        assert approvals.approvals_left == 0
        assert approvals.sufficient

    def test_reapprove(self):
        self.approvals.reapprove()
        self.api.call.assert_has_calls(
            [
                call(
                    POST(
                        endpoint="/projects/1/merge_requests/6/approve",
                        args={},
                        extract=None,
                    ),
                    sudo=1,
                ),
                call(
                    POST(
                        endpoint="/projects/1/merge_requests/6/approve",
                        args={},
                        extract=None,
                    ),
                    sudo=2,
                ),
            ]
        )

    @patch("marge.user.User.fetch_by_id")
    def test_get_reviewer_names_and_emails(self, user_fetch_by_id):
        user_fetch_by_id.side_effect = lambda id, _: marge.user.User(
            self.api, USERS[id]
        )
        assert _get_reviewer_names_and_emails(
            commits=[], approvals=self.approvals, api=self.api
        ) == ["Administrator <root@localhost>", "Roger Ebert <ebert@example.com>"]

    @patch("marge.user.User.fetch_by_id")
    def test_approvals_fails_when_same_author(self, user_fetch_by_id):
        info = dict(INFO, approved_by=list(INFO["approved_by"]))
        del info["approved_by"][1]
        approvals = Approvals(self.api, info)
        user_fetch_by_id.side_effect = lambda id, _: marge.user.User(
            self.api, USERS[id]
        )
        commits = [{"author_email": "root@localhost"}]
        with pytest.raises(CannotMerge):
            _get_reviewer_names_and_emails(
                commits=commits, approvals=approvals, api=self.api
            )

    @patch("marge.user.User.fetch_by_id")
    def test_approvals_succeeds_with_independent_author(self, user_fetch_by_id):
        user_fetch_by_id.side_effect = lambda id, _: marge.user.User(
            self.api, USERS[id]
        )
        print(INFO["approved_by"])
        commits = [{"author_email": "root@localhost"}]
        assert _get_reviewer_names_and_emails(
            commits=commits, approvals=self.approvals, api=self.api
        ) == ["Administrator <root@localhost>", "Roger Ebert <ebert@example.com>"]



import datetime

import pytest

import marge


@pytest.fixture
def tmp_repo(tmp_path_factory):
    source = str(tmp_path_factory.mktemp("source"))
    repo = marge.git.Repo(
        remote_url=source,
        local_path=str(tmp_path_factory.mktemp("checkout")),
        user_name="",
        user_email="",
        ssh_key_file=None,
        timeout=datetime.timedelta(seconds=10),
        reference=None,
        keep_committers=False,
    )
    repo.git("init", "-b", "main", source)
    yield repo


@pytest.fixture
def _git_user_config(monkeypatch):
    monkeypatch.setenv("GIT_AUTHOR_NAME", "Tester")
    monkeypatch.setenv("GIT_AUTHOR_EMAIL", "tester@example.com")
    monkeypatch.setenv("GIT_COMMITTER_NAME", "Tester")
    monkeypatch.setenv("GIT_COMMITTER_EMAIL", "tester@example.com")



# pylint: disable=too-many-locals,too-many-branches,too-many-statements
import datetime
import logging as log
import subprocess
import time
from typing import Optional

from . import approvals as mb_approvals
from . import git, gitlab
from . import job as mb_job
from . import merge_request as mb_merge_request
from . import project as mb_project
from . import user as mb_user
from .commit import Commit


class SingleMergeJob(mb_job.MergeJob):
    def __init__(
        self,
        *,
        api: gitlab.Api,
        user: mb_user.User,
        project: mb_project.Project,
        repo: git.Repo,
        options: mb_job.MergeJobOptions,
        merge_request: mb_merge_request.MergeRequest,
    ):
        super().__init__(
            api=api, user=user, project=project, repo=repo, options=options
        )
        self._merge_request = merge_request
        self._options = options

    def execute(self, exc_comment: Optional[str] = None) -> None:
        merge_request = self._merge_request

        log.info("Processing !%s - %r", merge_request.iid, merge_request.title)

        try:
            approvals = merge_request.fetch_approvals(self.approvals_factory)
            self.update_merge_request_and_accept(approvals)
            log.info("Successfully merged !%s.", merge_request.info["iid"])
        except mb_job.SkipMerge as err:
            log.warning("Skipping MR !%s: %s", merge_request.info["iid"], err.reason)
        except mb_job.CannotMerge as err:
            message = f"This branch couldn't be merged: {err.reason}"
            log.warning(message)
            self.unassign_from_mr(merge_request)
            merge_request.comment(message)
        except subprocess.TimeoutExpired as err:
            log.exception("TimeoutExpired", exc_info=err)
            comment = f"A git command timed out after {err.timeout} seconds"
            if exc_comment:
                comment += " " + exc_comment
            merge_request.comment(comment)
            self.unassign_from_mr(merge_request)
        except git.GitError as err:
            log.exception("Unexpected Git error", exc_info=err)
            comment = "Something seems broken on bot's local git repo. Check bot logs for more details."
            if exc_comment:
                comment += " " + exc_comment
            merge_request.comment(comment)
            raise
        except Exception as err:
            log.exception("Unexpected Exception", exc_info=err)
            comment = "Unexpected exception in bot while handling this MR."
            if exc_comment:
                comment += " " + exc_comment
            merge_request.comment(comment)
            self.unassign_from_mr(merge_request)
            raise

    def update_merge_request_and_accept(
        self, approvals: mb_approvals.Approvals
    ) -> None:
        api = self._api
        merge_request = self._merge_request
        updated_into_up_to_date_target_branch = False

        while not updated_into_up_to_date_target_branch:
            self.ensure_mergeable_mr(merge_request)
            source_project, source_repo_url, _ = self.fetch_source_project(
                merge_request
            )
            target_project = self.get_target_project(merge_request)

            # NB. this will be a no-op if there is nothing to update/rewrite
            target_sha, _updated_sha, actual_sha = (
                self.update_from_target_branch_and_push(
                    merge_request, source_repo_url=source_repo_url
                )
            )

            if _updated_sha == actual_sha and self._options.guarantee_final_pipeline:
                log.info("No commits on target branch to fuse, triggering pipeline...")
                pipeline_info = merge_request.trigger_pipeline()

                log.info(
                    "Pipeline %s is triggered, waiting for it to finish...",
                    pipeline_info.get("id"),
                )
                self.wait_for_ci_to_pass(merge_request, actual_sha)

            log.info(
                "Commit id to merge %r into: %r (updated sha: %r)",
                actual_sha,
                target_sha,
                _updated_sha,
            )
            time.sleep(5)

            sha_now = Commit.last_on_branch(
                source_project.id, merge_request.source_branch, api
            ).id
            # Make sure no-one managed to race and push to the branch in the
            # meantime, because we're about to impersonate the approvers, and
            # we don't want to approve unreviewed commits
            if sha_now != actual_sha:
                raise mb_job.CannotMerge(
                    "Someone pushed to branch while we were trying to merge"
                )

            self.maybe_reapprove(merge_request, approvals)

            if target_project.only_allow_merge_if_pipeline_succeeds:
                self.wait_for_ci_to_pass(merge_request, actual_sha)
                time.sleep(2)

            try:
                self.wait_for_merge_status_to_resolve(merge_request)
            except mb_job.NeedsRebase:
                continue

            self.ensure_mergeable_mr(merge_request)

            try:
                ret = merge_request.accept(
                    remove_branch=merge_request.force_remove_source_branch,
                    sha=actual_sha,
                    merge_when_pipeline_succeeds=bool(
                        target_project.only_allow_merge_if_pipeline_succeeds
                    ),
                    squash_on_merge=merge_request.squash_on_merge,
                )
                log.debug("merge_request.accept result: %s", ret)
            except gitlab.Unprocessable as err:
                new_target_sha = Commit.last_on_branch(
                    self._project.id, merge_request.target_branch, api
                ).id
                # target_branch has moved under us since we updated, just try again
                if new_target_sha != target_sha:
                    log.info("Someone was naughty and by-passed marge")
                    merge_request.comment(
                        "A merge request may have skipped the queue, trying again..."
                    )
                    continue
                # otherwise the source branch has been pushed to or something
                # unexpected went wrong in either case, we expect the user to
                # explicitly re-assign to marge (after resolving potential
                # problems)
                raise mb_job.CannotMerge(
                    f"Merge request was rejected by GitLab: {err.error_message!r}"
                ) from err
            except gitlab.Unauthorized as err:
                log.warning("Unauthorized!")
                raise mb_job.CannotMerge(
                    "This user does not have permission to merge MRs."
                ) from err
            except gitlab.NotFound as ex:
                log.warning("Not Found!: %s", ex)
                merge_request.refetch_info()
                if merge_request.state == "merged":
                    # someone must have hit "merge when build succeeds" and we lost the race,
                    # the branch is gone and we got a 404. Anyway, our job here is done.
                    # (see #33)
                    updated_into_up_to_date_target_branch = True
                else:
                    log.warning(
                        "For the record, merge request state is %r", merge_request.state
                    )
                    raise
            except gitlab.MethodNotAllowed as ex:
                log.warning("Not Allowed!: %s", ex)
                merge_request.refetch_info()
                if merge_request.draft:
                    raise mb_job.CannotMerge(
                        "The request was marked as Draft as the bot was processing it."
                        " Maybe a Draft commit was pushed?"
                    ) from ex
                if merge_request.state == "reopened":
                    raise mb_job.CannotMerge(
                        "GitLab could not merge this branch. Possibly a Push Rule or a git-hook "
                        "is rejecting the commits; maybe the bot's email needs to be allow-listed?"
                    ) from ex
                if merge_request.state == "closed":
                    raise mb_job.CannotMerge(
                        "The merge request was closed while the merge was being attempted."
                    ) from ex
                if merge_request.state == "merged":
                    # We are not covering any observed behaviour here, but if at this
                    # point the request is merged, our job is done, so no need to complain
                    log.info("Merge request is already merged, someone was faster!")
                    updated_into_up_to_date_target_branch = True
                else:
                    raise mb_job.CannotMerge(
                        "Gitlab could not merge this branch for an unknown reason."
                        + (
                            " Maybe there are unresolved discussions?"
                            if self._project.only_allow_merge_if_all_discussions_are_resolved
                            else ""
                        )
                    ) from ex
            except gitlab.ApiError as err:
                log.exception("Unanticipated ApiError from GitLab on merge attempt")
                raise mb_job.CannotMerge(
                    "The was an unknown GitLab error. Check bot logs for more details."
                ) from err
            else:
                self.wait_for_branch_to_be_merged()
                updated_into_up_to_date_target_branch = True

    def wait_for_branch_to_be_merged(self) -> None:
        merge_request = self._merge_request
        time_0 = datetime.datetime.utcnow()
        waiting_time_in_secs = 10

        while datetime.datetime.utcnow() - time_0 < self._merge_timeout:
            merge_request.refetch_info()

            if merge_request.state == "merged":
                return  # success!
            if merge_request.state == "closed":
                raise mb_job.CannotMerge(
                    "The merge request was closed while the merge was being attempted."
                )
            assert merge_request.state in (
                "opened",
                "reopened",
                "locked",
            ), merge_request.state

            log.info(
                "Giving %s more secs for !%s to be merged...",
                waiting_time_in_secs,
                merge_request.iid,
            )
            time.sleep(waiting_time_in_secs)

        raise mb_job.CannotMerge(
            "It is taking too long to see the request marked as merged!"
        )



from unittest.mock import Mock, call

import pytest

import marge.user
from marge.gitlab import GET, POST, PUT, Api, BadRequest
from marge.merge_request import (
    NO_JOBS_MESSAGE,
    RULES_PREVENT_JOBS_MESSAGE,
    MergeRequest,
    MergeRequestRebaseFailed,
)
from tests.test_user import INFO as USER_INFO

_MARGE_ID = 77

INFO = {
    "id": 42,
    "iid": 54,
    "title": "a title",
    "project_id": 1234,
    "assignees": [{"id": _MARGE_ID}],
    "author": {"id": 88},
    "state": "opened",
    "sha": "dead4g00d",
    "source_project_id": 5678,
    "target_project_id": 1234,
    "source_branch": "useless_new_feature",
    "force_remove_source_branch": True,
    "target_branch": "master",
    "draft": False,
}

DISCUSSION = {
    "id": "aabbcc0044",
    "notes": [
        {
            "id": 12,
            "body": "assigned to @john_smith",
            "created_at": "2020-08-04T06:56:11.854Z",
        },
        {
            "id": 13,
            "body": "assigned to @john_smith",
            "created_at": "2020-08-18T06:52:58.093Z",
        },
    ],
}


# pylint: disable=attribute-defined-outside-init
class TestMergeRequest:
    def setup_method(self, _method):
        self.api = Mock(Api)
        self.merge_request = MergeRequest(api=self.api, info=INFO)

    def test_fetch_by_iid(self):
        api = self.api
        api.call = Mock(return_value=INFO)

        merge_request = MergeRequest.fetch_by_iid(
            project_id=1234, merge_request_iid=54, api=api
        )

        api.call.assert_called_once_with(
            GET(
                "/projects/1234/merge_requests/54",
                {"include_rebase_in_progress": "true"},
            )
        )
        assert merge_request.info == INFO

    def test_fetch_all_mine(self):
        prj1, prj2 = INFO, dict(INFO, id=678)

        api = self.api
        api.collect_all_pages = Mock(return_value=[prj1, prj2])
        user = marge.user.User(api=None, info=dict(USER_INFO, id=_MARGE_ID))

        result = MergeRequest.fetch_all_mine(api, user, "created_at")
        api.collect_all_pages.assert_called_once_with(
            GET(
                "/merge_requests",
                {
                    "scope": "assigned_to_me",
                    "state": "opened",
                    "order_by": "created_at",
                    "sort": "asc",
                },
            )
        )
        assert [prj.info for prj in result] == [prj1, prj2]

    def test_refetch_info(self):
        new_info = dict(INFO, state="closed")
        self.api.call = Mock(return_value=new_info)

        self.merge_request.refetch_info()
        self.api.call.assert_called_once_with(
            GET(
                "/projects/1234/merge_requests/54",
                {"include_rebase_in_progress": "true"},
            )
        )
        assert self.merge_request.info == new_info

    def test_properties(self):
        assert self.merge_request.id == 42
        assert self.merge_request.project_id == 1234
        assert self.merge_request.iid == 54
        assert self.merge_request.title == "a title"
        assert self.merge_request.assignee_ids == [77]
        assert self.merge_request.author_id == 88
        assert self.merge_request.state == "opened"
        assert self.merge_request.source_branch == "useless_new_feature"
        assert self.merge_request.target_branch == "master"
        assert self.merge_request.sha == "dead4g00d"
        assert self.merge_request.source_project_id == 5678
        assert self.merge_request.target_project_id == 1234
        assert self.merge_request.draft is False

        self._load({"assignees": []})
        assert self.merge_request.assignee_ids == []

    def test_comment(self):
        self.merge_request.comment("blah")
        self.api.call.assert_called_once_with(
            POST("/projects/1234/merge_requests/54/notes", {"body": "blah"})
        )

    def test_assign(self):
        self.merge_request.assign_to(42)
        self.api.call.assert_called_once_with(
            PUT("/projects/1234/merge_requests/54", {"assignee_id": 42})
        )

    def test_unassign(self):
        self.merge_request.unassign()
        self.api.call.assert_called_once_with(
            PUT("/projects/1234/merge_requests/54", {"assignee_id": 0})
        )

    def test_rebase_was_not_in_progress_no_error(self):
        expected = [
            (
                GET(
                    "/projects/1234/merge_requests/54",
                    {"include_rebase_in_progress": "true"},
                ),  # refetch_info -> not in progress
                INFO,
            ),
            (PUT("/projects/1234/merge_requests/54/rebase"), True),
            (
                GET(
                    "/projects/1234/merge_requests/54",
                    {"include_rebase_in_progress": "true"},
                ),  # refetch_info -> in progress
                dict(INFO, rebase_in_progress=True),
            ),
            (
                GET(
                    "/projects/1234/merge_requests/54",
                    {"include_rebase_in_progress": "true"},
                ),  # refetch_info -> succeeded
                dict(INFO, rebase_in_progress=False),
            ),
        ]

        self.api.call = Mock(side_effect=[resp for (req, resp) in expected])
        self.merge_request.rebase()
        self.api.call.assert_has_calls([call(req) for (req, resp) in expected])

    def test_rebase_was_not_in_progress_error(self):
        expected = [
            (
                GET(
                    "/projects/1234/merge_requests/54",
                    {"include_rebase_in_progress": "true"},
                ),  # refetch_info -> not in progress
                INFO,
            ),
            (PUT("/projects/1234/merge_requests/54/rebase"), True),
            (
                GET(
                    "/projects/1234/merge_requests/54",
                    {"include_rebase_in_progress": "true"},
                ),  # refetch_info -> BOOM
                dict(
                    INFO,
                    rebase_in_progress=False,
                    merge_error="Rebase failed. Please rebase locally",
                ),
            ),
        ]

        self.api.call = Mock(side_effect=[resp for (req, resp) in expected])

        with pytest.raises(MergeRequestRebaseFailed):
            self.merge_request.rebase()
        self.api.call.assert_has_calls([call(req) for (req, resp) in expected])

    def test_rebase_was_in_progress_no_error(self):
        expected = [
            (
                GET(
                    "/projects/1234/merge_requests/54",
                    {"include_rebase_in_progress": "true"},
                ),  # refetch_info -> in progress
                dict(INFO, rebase_in_progress=True),
            ),
            (
                GET(
                    "/projects/1234/merge_requests/54",
                    {"include_rebase_in_progress": "true"},
                ),  # refetch_info -> in progress
                dict(INFO, rebase_in_progress=True),
            ),
            (
                GET(
                    "/projects/1234/merge_requests/54",
                    {"include_rebase_in_progress": "true"},
                ),  # refetch_info -> succeeded
                dict(INFO, rebase_in_progress=False),
            ),
        ]
        self.api.call = Mock(side_effect=[resp for (req, resp) in expected])
        self.merge_request.rebase()
        self.api.call.assert_has_calls([call(req) for (req, resp) in expected])

    @pytest.mark.parametrize("squash_on_merge", [True, False])
    def test_accept_auto_squash_is_boolean(self, squash_on_merge):
        self._load(dict(INFO, sha="badc0de"))
        self.merge_request.accept(squash_on_merge=squash_on_merge)
        self.api.call.assert_called_once_with(
            PUT(
                "/projects/1234/merge_requests/54/merge",
                {
                    "merge_when_pipeline_succeeds": True,
                    "should_remove_source_branch": False,
                    "sha": "badc0de",
                    "squash": squash_on_merge,
                },
            )
        )

    def test_accept_remove_branch(self):
        self._load(dict(INFO, sha="badc0de"))

        for boolean in (True, False):
            self.merge_request.accept(remove_branch=boolean)
            self.api.call.assert_called_once_with(
                PUT(
                    "/projects/1234/merge_requests/54/merge",
                    {
                        "merge_when_pipeline_succeeds": True,
                        "should_remove_source_branch": boolean,
                        "sha": "badc0de",
                        "squash": False,
                    },
                )
            )
            self.api.call.reset_mock()

    def test_accept_sha(self):
        self._load(dict(INFO, sha="badc0de"))
        self.merge_request.accept(sha="g00dc0de")
        self.api.call.assert_called_once_with(
            PUT(
                "/projects/1234/merge_requests/54/merge",
                {
                    "merge_when_pipeline_succeeds": True,
                    "should_remove_source_branch": False,
                    "sha": "g00dc0de",
                    "squash": False,
                },
            )
        )

    def test_accept_merge_when_pipeline_succeeds(self):
        self._load(dict(INFO, sha="badc0de"))
        self.merge_request.accept(merge_when_pipeline_succeeds=False)
        self.api.call.assert_called_once_with(
            PUT(
                "/projects/1234/merge_requests/54/merge",
                {
                    "merge_when_pipeline_succeeds": False,
                    "should_remove_source_branch": False,
                    "sha": "badc0de",
                    "squash": False,
                },
            )
        )

    def test_fetch_assigned_at(self):
        api = self.api
        dis1, dis2 = DISCUSSION, dict(DISCUSSION, id=679)
        mr1 = INFO
        user = marge.user.User(api=None, info=dict(USER_INFO, id=_MARGE_ID))
        api.collect_all_pages = Mock(return_value=[dis1, dis2])
        result = MergeRequest.fetch_assigned_at(user=user, api=api, merge_request=mr1)
        api.collect_all_pages.assert_called_once_with(
            GET("/projects/1234/merge_requests/54/discussions")
        )
        assert result == 1597733578.093

    def test_trigger_pipeline_succeeds(self):
        api = self.api
        expected_result = object()

        def side_effect(request):
            if request.endpoint == "/projects/1234/merge_requests/54/pipelines":
                return expected_result
            return None

        api.call = Mock(side_effect=side_effect)

        result = self.merge_request.trigger_pipeline()

        assert api.call.call_args_list == [
            call(POST("/projects/1234/merge_requests/54/pipelines"))
        ]

        assert result == expected_result

    def test_trigger_pipeline_fallback_succeeds(self):
        api = self.api
        expected_result = object()

        def side_effect(request):
            if request.endpoint == "/projects/1234/merge_requests/54/pipelines":
                raise BadRequest(NO_JOBS_MESSAGE, 400)
            if request.endpoint == "/projects/1234/pipeline?ref=useless_new_feature":
                return expected_result
            return None

        api.call = Mock(side_effect=side_effect)

        result = self.merge_request.trigger_pipeline()

        assert api.call.call_args_list == [
            call(POST("/projects/1234/merge_requests/54/pipelines")),
            call(POST("/projects/1234/pipeline?ref=useless_new_feature")),
        ]

        assert result == expected_result

    def test_trigger_pipeline_rules_fallback_succeeds(self):
        api = self.api
        expected_result = object()

        def side_effect(request):
            if request.endpoint == "/projects/1234/merge_requests/54/pipelines":
                raise BadRequest(RULES_PREVENT_JOBS_MESSAGE, 400)
            if request.endpoint == "/projects/1234/pipeline?ref=useless_new_feature":
                return expected_result
            return None

        api.call = Mock(side_effect=side_effect)

        result = self.merge_request.trigger_pipeline()

        assert api.call.call_args_list == [
            call(POST("/projects/1234/merge_requests/54/pipelines")),
            call(POST("/projects/1234/pipeline?ref=useless_new_feature")),
        ]

        assert result == expected_result

    def test_trigger_pipeline_rules_fallback_nested_succeeds(self):
        api = self.api
        expected_result = object()

        def side_effect(request):
            if request.endpoint == "/projects/1234/merge_requests/54/pipelines":
                raise BadRequest(RULES_PREVENT_JOBS_MESSAGE, 400)
            if request.endpoint == "/projects/1234/pipeline?ref=useless_new_feature":
                return expected_result
            return None

        api.call = Mock(side_effect=side_effect)

        result = self.merge_request.trigger_pipeline()

        assert api.call.call_args_list == [
            call(POST("/projects/1234/merge_requests/54/pipelines")),
            call(POST("/projects/1234/pipeline?ref=useless_new_feature")),
        ]

        assert result == expected_result

    def test_trigger_pipeline_fallback_fails(self):
        api = self.api

        def side_effect(request):
            if request.endpoint == "/projects/1234/merge_requests/54/pipelines":
                raise BadRequest("Another error.", 500)

        api.call = Mock(side_effect=side_effect)

        with pytest.raises(BadRequest):
            self.merge_request.trigger_pipeline()

        assert api.call.call_args_list == [
            call(POST("/projects/1234/merge_requests/54/pipelines"))
        ]

    def _load(self, json):
        old_mock = self.api.call
        self.api.call = Mock(return_value=json)
        self.merge_request.refetch_info()
        self.api.call.assert_called_with(
            GET(
                "/projects/1234/merge_requests/54",
                {"include_rebase_in_progress": "true"},
            )
        )
        self.api.call = old_mock



import urllib.parse
from typing import TYPE_CHECKING

from . import gitlab

GET = gitlab.GET
DELETE = gitlab.DELETE


class Branch(gitlab.Resource):
    @classmethod
    def fetch_by_name(cls, project_id: int, branch: str, api: gitlab.Api) -> "Branch":
        info = api.call(GET(f"/projects/{project_id}/repository/branches/{branch}"))
        if TYPE_CHECKING:
            assert isinstance(info, dict)
        return cls(api, info)

    @classmethod
    def delete_by_name(cls, project_id: int, branch: str, api: gitlab.Api) -> None:
        api.call(
            DELETE(
                f"/projects/{project_id}/repository/branches/"
                f'{urllib.parse.quote(branch, safe="")}'
            )
        )

    @property
    def id(self) -> int:
        raise NotImplementedError()

    @property
    def name(self) -> str:
        name = self.info["name"]
        if TYPE_CHECKING:
            assert isinstance(name, str)
        return name

    @property
    def protected(self) -> bool:
        protected = self.info["protected"]
        if TYPE_CHECKING:
            assert isinstance(protected, bool)
        return protected



import asyncio
import importlib.util
import inspect
import logging as log
from pathlib import Path
from typing import Any, Callable, Optional, cast

HookType = Callable[..., Optional[Any]]


def load_hook(hooks_dir_path: Path, hook_name: str) -> Optional[HookType]:
    file_path = hooks_dir_path / f"{hook_name}.py"
    if not file_path.is_file():
        log.info("Hook file not found: %s", file_path)
        return None

    try:
        spec = importlib.util.spec_from_file_location(hook_name, file_path)
        if spec is None or spec.loader is None:
            log.warning("Could not load spec for file: %s", file_path)
            return None

        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return cast(HookType, module.main)

    except (AttributeError, RuntimeError, TypeError, ValueError) as err:
        log.error("Error loading hook %s: %s", hook_name, err, exc_info=True)

    return None


def run_hook(
    hooks_dir: str, hook_name: str, *args: Any, **kwargs: Any
) -> Optional[Any]:
    hooks_dir_path = Path(hooks_dir)
    if not hooks_dir_path.is_dir():
        log.info("Hook directory not found: %s", hooks_dir)
        return None

    hook = load_hook(hooks_dir_path, hook_name)
    if hook is None:
        log.info("Hook '%s' not loaded.", hook_name)
        return None

    try:
        if inspect.iscoroutinefunction(hook):
            return asyncio.run(hook(*args, **kwargs))
        return hook(*args, **kwargs)
    except (AttributeError, RuntimeError, TypeError, ValueError) as err:
        log.error("Error running hook '%s': %s", hook_name, err, exc_info=True)
        return None



import abc
import dataclasses
import datetime
import tempfile
from typing import TYPE_CHECKING, Optional

from . import git
from . import project as mb_project
from . import user as mb_user


@dataclasses.dataclass
class RepoManager(abc.ABC):
    user: mb_user.User
    root_dir: str
    timeout: Optional[datetime.timedelta] = None
    reference: Optional[str] = None
    keep_committers: bool = False

    def __post_init__(self) -> None:
        self._repos: dict[int, git.Repo] = {}

    @abc.abstractmethod
    def _get_project_url(self, project: mb_project.Project) -> str: ...

    @abc.abstractmethod
    def _git_repo(self, repo_url: str, local_repo_dir: str) -> git.Repo: ...

    def repo_for_project(self, project: mb_project.Project) -> git.Repo:
        repo = self._repos.get(project.id)
        if not repo or repo.remote_url != self._get_project_url(project):
            repo_url = self._get_project_url(project)
            local_repo_dir = tempfile.mkdtemp(dir=self.root_dir)

            repo = self._git_repo(repo_url, local_repo_dir)
            repo.clone()
            repo.configure()

            self._repos[project.id] = repo

        return repo


@dataclasses.dataclass
class SshRepoManager(RepoManager):
    ssh_key_file: Optional[str] = None

    def _get_project_url(self, project: mb_project.Project) -> str:
        return project.ssh_url_to_repo

    def _git_repo(self, repo_url: str, local_repo_dir: str) -> git.Repo:
        if TYPE_CHECKING:
            assert self.user.email is not None
        return git.Repo(
            repo_url,
            local_repo_dir,
            self.user.email,
            self.user.name,
            ssh_key_file=self.ssh_key_file,
            timeout=self.timeout,
            reference=self.reference,
            keep_committers=self.keep_committers,
        )


@dataclasses.dataclass
class HttpsRepoManager(RepoManager):
    auth_token: Optional[str] = None

    def _get_project_url(self, project: mb_project.Project) -> str:
        return project.http_url_to_repo

    def _git_repo(self, repo_url: str, local_repo_dir: str) -> git.Repo:
        if TYPE_CHECKING:
            assert self.auth_token is not None
            assert self.user.email is not None
        return git.Repo(
            repo_url,
            local_repo_dir,
            self.user.email,
            self.user.name,
            ssh_key_file=None,
            timeout=self.timeout,
            reference=self.reference,
            auth_token=self.auth_token,
            keep_committers=self.keep_committers,
        )



import collections
import dataclasses
import datetime
import logging as log
import re
import time
from tempfile import TemporaryDirectory
from typing import Optional

from . import batch_job, git, gitlab, job
from . import merge_request as merge_request_module
from . import single_merge_job, store
from . import user as user_module
from .project import Project

MergeRequest = merge_request_module.MergeRequest


class Bot:
    def __init__(self, *, api: gitlab.Api, config: "BotConfig"):
        self._api = api
        self._config = config

        user = config.user
        opts = config.merge_opts

        if not user.is_admin:
            assert (
                not opts.reapprove
            ), f"{user.username} is not an admin, can't impersonate!"
            assert (
                not opts.add_reviewers
            ), f"{user.username} is not an admin, can't lookup Reviewed-by: email addresses "

    def start(self) -> None:
        with TemporaryDirectory() as root_dir:
            repo_manager: store.RepoManager
            if self._config.use_https:
                repo_manager = store.HttpsRepoManager(
                    user=self.user,
                    root_dir=root_dir,
                    auth_token=self._config.auth_token,
                    timeout=self._config.git_timeout,
                    reference=self._config.git_reference_repo,
                    keep_committers=self._config.merge_opts.keep_committers,
                )
            else:
                repo_manager = store.SshRepoManager(
                    user=self.user,
                    root_dir=root_dir,
                    ssh_key_file=self._config.ssh_key_file,
                    timeout=self._config.git_timeout,
                    reference=self._config.git_reference_repo,
                    keep_committers=self._config.merge_opts.keep_committers,
                )
            self._run(repo_manager)

    @property
    def user(self) -> user_module.User:
        return self._config.user

    @property
    def api(self) -> gitlab.Api:
        return self._api

    def _run(self, repo_manager: store.RepoManager) -> None:
        sleep_time_between_iterations = 30
        while True:
            pending_merge_requests = self._get_assigned_merge_requests()
            for project, merge_requests in pending_merge_requests.items():
                self._process_merge_requests(repo_manager, project, merge_requests)
            if self._config.cli:
                return

            log.info("Sleeping for %s seconds...", sleep_time_between_iterations)
            time.sleep(sleep_time_between_iterations)

    def _get_assigned_merge_requests(self) -> dict[Project, list[MergeRequest]]:
        log.info("Fetching merge requests assigned to me...")
        project_merge_requests = collections.defaultdict(list)
        branch_regexp = self._config.branch_regexp
        source_branch_regexp = self._config.source_branch_regexp
        assigned_merge_requests = MergeRequest.fetch_all_mine(
            self._api, self.user, self._config.merge_order
        )
        for merge_request in assigned_merge_requests:
            project_id = merge_request.project_id
            if not branch_regexp.match(merge_request.target_branch):
                log.debug("MR does not match branch_regexp: %s", merge_request.web_url)
                continue

            if not source_branch_regexp.match(merge_request.source_branch):
                log.debug(
                    "MR does not match source_branch_regexp: %s", merge_request.web_url
                )
                continue

            project_merge_requests[project_id].append(merge_request)

        filtered_merge_requests = {}
        project_regexp = self._config.project_regexp
        for project_id, merge_requests in project_merge_requests.items():
            project = Project.fetch_by_id(project_id, self.api)

            if project.archived:
                log.debug("Project is archived: %s", project.path_with_namespace)
                continue

            if project_regexp.match(project.path_with_namespace):
                filtered_merge_requests[project] = merge_requests
            else:
                log.debug(
                    "Project does not match project_regexp: %s",
                    project.path_with_namespace,
                )

        return filtered_merge_requests

    def _process_merge_requests(
        self,
        repo_manager: store.RepoManager,
        project: Project,
        merge_requests: list[MergeRequest],
    ) -> None:
        if not merge_requests:
            log.info("Nothing to merge at this point...")
            return

        try:
            repo = repo_manager.repo_for_project(project)
        except git.GitError:
            log.exception("Couldn't initialize repository for project!")
            raise

        log.info("Got %s requests to merge;", len(merge_requests))
        if self._config.batch and len(merge_requests) > 1:
            log.info(
                "Attempting to merge as many MRs as possible using BatchMergeJob..."
            )
            batch_merge_job = batch_job.BatchMergeJob(
                api=self._api,
                user=self.user,
                project=project,
                merge_requests=merge_requests,
                repo=repo,
                options=self._config.merge_opts,
                batch_branch_name=self._config.batch_branch_name,
            )
            try:
                batch_merge_job.execute(exc_comment=self._config.exc_comment)
                return
            except batch_job.CannotBatch as err:
                log.warning("BatchMergeJob aborted: %s", err)
            except job.CannotMerge as err:
                log.warning("BatchMergeJob failed: %s", err)
                return
            except git.GitError as err:
                log.exception("BatchMergeJob failed: %s", err)
        log.info("Attempting to merge the oldest MR...")
        merge_request = merge_requests[0]
        merge_job = self._get_single_job(
            project=project,
            merge_request=merge_request,
            repo=repo,
            options=self._config.merge_opts,
        )
        merge_job.execute(exc_comment=self._config.exc_comment)

    def _get_single_job(
        self,
        project: Project,
        merge_request: MergeRequest,
        repo: git.Repo,
        options: job.MergeJobOptions,
    ) -> single_merge_job.SingleMergeJob:
        return single_merge_job.SingleMergeJob(
            api=self._api,
            user=self.user,
            project=project,
            merge_request=merge_request,
            repo=repo,
            options=options,
        )


@dataclasses.dataclass
class BotConfig:
    user: user_module.User
    use_https: bool
    auth_token: str
    ssh_key_file: Optional[str]
    project_regexp: "re.Pattern[str]"
    merge_order: str
    merge_opts: job.MergeJobOptions
    git_timeout: datetime.timedelta
    git_reference_repo: str
    branch_regexp: "re.Pattern[str]"
    source_branch_regexp: "re.Pattern[str]"
    batch: bool
    cli: bool
    batch_branch_name: str
    exc_comment: Optional[str]


MergeJobOptions = job.MergeJobOptions
Fusion = job.Fusion



import os.path
from unittest import mock

import pytest

import marge.git
import marge.store
import marge.user
from tests.test_git import get_calls as get_git_calls
from tests.test_project import INFO as PRJ_INFO
from tests.test_user import INFO as USER_INFO


# pylint: disable=attribute-defined-outside-init
@mock.patch("marge.git._run")
class TestRepoManager:
    @pytest.fixture
    def user(self):
        return marge.user.User(
            api=None,
            info=dict(USER_INFO, name="Peter Parker", email="pparker@bugle.com"),
        )

    @pytest.fixture
    def ssh_repo_manager(self, user, tmp_path):
        return marge.store.SshRepoManager(
            user=user, root_dir=str(tmp_path), ssh_key_file="/ssh/key"
        )

    @pytest.fixture
    def https_repo_manager(self, user, tmp_path):
        return marge.store.HttpsRepoManager(
            user=user, root_dir=str(tmp_path), auth_token="glpat-TEST"
        )

    def new_project(self, project_id, path_with_namespace):
        schemeless_url_to_repo = f"buh.com/{path_with_namespace}.git"
        info = dict(
            PRJ_INFO,
            id=project_id,
            path_with_namespace=path_with_namespace,
            ssh_url_to_repo=f"ssh://{schemeless_url_to_repo}",
            http_url_to_repo=f"https://{schemeless_url_to_repo}",
        )
        return marge.project.Project(api=None, info=info)

    def test_creates_and_initializes_repo(self, git_run, ssh_repo_manager):
        project = self.new_project(1234, "some/stuff")

        git_run.assert_not_called()

        repo = ssh_repo_manager.repo_for_project(project)

        assert os.path.dirname(repo.local_path) == ssh_repo_manager.root_dir
        assert repo.local_path != ssh_repo_manager.root_dir

        env = (
            f"GIT_SSH_COMMAND='{marge.git.GIT_SSH_COMMAND} -F /dev/null "
            f"-o IdentitiesOnly=yes -i /ssh/key'"
        )
        assert get_git_calls(git_run) == [
            f"{env} git clone --origin=origin --filter=blob:none {project.ssh_url_to_repo} "
            f"{repo.local_path}",
            f"{env} git -C {repo.local_path} config user.email pparker@bugle.com",
            f"{env} git -C {repo.local_path} config user.name 'Peter Parker'",
            f"{env} git -C {repo.local_path} config gpg.format ssh",
            f"{env} git -C {repo.local_path} config user.signingKey /ssh/key",
        ]

    def test_caches_repos_by_id(self, git_run, ssh_repo_manager):
        project = self.new_project(1234, "some/stuff")
        same_project = marge.project.Project(
            api=None, info=dict(project.info, name="same/stuff")
        )

        assert git_run.call_count == 0

        repo_first_call = ssh_repo_manager.repo_for_project(project)
        assert git_run.call_count == 5

        repo_second_call = ssh_repo_manager.repo_for_project(same_project)
        assert repo_second_call is repo_first_call
        assert git_run.call_count == 5

    def test_stops_caching_if_ssh_url_changed(self, git_run, ssh_repo_manager):
        project = self.new_project(1234, "some/stuff")

        assert git_run.call_count == 0

        repo_first_call = ssh_repo_manager.repo_for_project(project)
        assert git_run.call_count == 5

        different_ssh_url = self.new_project(1234, "same/stuff")

        repo_second_call = ssh_repo_manager.repo_for_project(different_ssh_url)
        assert git_run.call_count == 10
        assert (
            repo_first_call.remote_url
            != repo_second_call.remote_url
            == different_ssh_url.ssh_url_to_repo
        )

    def test_handles_different_projects(self, git_run, ssh_repo_manager):
        project_1 = self.new_project(1234, "some/stuff")
        project_2 = self.new_project(5678, "other/things")

        assert git_run.call_count == 0

        repo_1 = ssh_repo_manager.repo_for_project(project_1)
        assert git_run.call_count == 5

        repo_2 = ssh_repo_manager.repo_for_project(project_2)
        assert git_run.call_count == 10

        assert repo_1.local_path != repo_2.local_path

    def test_no_dups_with_token(self, git_run, https_repo_manager):
        project = self.new_project(1234, "some/stuff")

        assert git_run.call_count == 0

        repo_initial = https_repo_manager.repo_for_project(project)
        assert git_run.call_count == 3

        repo_repeat = https_repo_manager.repo_for_project(project)
        # Call count should not have changed as the repo is cached.
        assert git_run.call_count == 3

        assert repo_initial.local_path == repo_repeat.local_path



import datetime
import logging as log
import time
from typing import TYPE_CHECKING, Any, Callable, Optional

from . import gitlab
from . import user as mb_user
from .approvals import Approvals

NO_JOBS_MESSAGE = "No stages / jobs for this pipeline."
RULES_PREVENT_JOBS_MESSAGE = (
    "Pipeline will not run for the selected trigger. "
    "The rules configuration prevented any jobs from being added to the pipeline."
)


class MergeRequest(gitlab.Resource):
    # pylint: disable=too-many-public-methods
    @classmethod
    def create(
        cls, api: gitlab.Api, project_id: int, params: dict[str, Any]
    ) -> "MergeRequest":
        merge_request_info = api.call(
            gitlab.POST(f"/projects/{project_id}/merge_requests", params)
        )
        if TYPE_CHECKING:
            assert isinstance(merge_request_info, dict)
        return cls(api, merge_request_info)

    @classmethod
    def search(
        cls, api: gitlab.Api, project_id: int, params: dict[str, Any]
    ) -> list["MergeRequest"]:
        merge_requests = api.collect_all_pages(
            gitlab.GET(f"/projects/{project_id}/merge_requests", params)
        )
        return [cls(api, merge_request) for merge_request in merge_requests]

    @classmethod
    def fetch_by_iid(
        cls, project_id: int, merge_request_iid: int, api: gitlab.Api
    ) -> "MergeRequest":
        merge_request = cls(api, {"iid": merge_request_iid, "project_id": project_id})
        merge_request.refetch_info()
        return merge_request

    @classmethod
    def fetch_assigned_at(
        cls, user: mb_user.User, api: gitlab.Api, merge_request: dict[str, Any]
    ) -> float:
        assigned_at = 0.0
        all_discussions = api.collect_all_pages(
            gitlab.GET(
                f'/projects/{merge_request.get("project_id")}/'
                + f'merge_requests/{merge_request.get("iid")}/discussions'
            )
        )
        match_body = f"assigned to @{user.username}"
        for discussion in all_discussions:
            for note in discussion.get("notes", []):
                if match_body in note.get("body"):
                    date_string = note.get("created_at")
                    date_format = "%Y-%m-%dT%H:%M:%S.%f%z"
                    assigned = datetime.datetime.strptime(
                        date_string, date_format
                    ).timestamp()
                    assigned_at = max(assigned_at, assigned)
        return assigned_at

    @classmethod
    def fetch_all_mine(
        cls, api: gitlab.Api, user: mb_user.User, merge_order: str
    ) -> list["MergeRequest"]:
        request_merge_order = (
            "created_at" if merge_order == "assigned_at" else merge_order
        )

        assigned_mrs = api.collect_all_pages(
            gitlab.GET(
                "/merge_requests",
                {
                    "scope": "assigned_to_me",
                    "state": "opened",
                    "order_by": request_merge_order,
                    "sort": "asc",
                },
            )
        )

        if merge_order == "assigned_at":
            assigned_mrs.sort(key=lambda amr: cls.fetch_assigned_at(user, api, amr))

        return [cls(api, merge_request_info) for merge_request_info in assigned_mrs]

    @property
    def id(self) -> int:
        result = self._info["id"]
        if TYPE_CHECKING:
            assert isinstance(result, int)
        return result

    @property
    def project_id(self) -> int:
        result = self.info["project_id"]
        if TYPE_CHECKING:
            assert isinstance(result, int)
        return result

    @property
    def iid(self) -> int:
        result = self.info["iid"]
        if TYPE_CHECKING:
            assert isinstance(result, int)
        return result

    @property
    def title(self) -> str:
        result = self.info["title"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def state(self) -> str:
        result = self.info["state"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def detailed_merge_status(self) -> str:
        result = self.info["detailed_merge_status"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def rebase_in_progress(self) -> bool:
        result = self.info.get("rebase_in_progress", False)
        if TYPE_CHECKING:
            assert isinstance(result, bool)
        return result

    @property
    def merge_error(self) -> Optional[str]:
        result = self.info.get("merge_error")
        if TYPE_CHECKING:
            if result is not None:
                assert isinstance(result, str)
        return result

    @property
    def assignee_ids(self) -> list[Optional[int]]:
        if "assignees" in self.info:
            return [assignee.get("id") for assignee in (self.info["assignees"] or [])]
        return [(self.info.get("assignee", {}) or {}).get("id")]

    @property
    def author_id(self) -> Optional[int]:
        author = self.info["author"]
        if TYPE_CHECKING:
            assert isinstance(author, dict)
        return author.get("id")

    @property
    def source_branch(self) -> str:
        result = self.info["source_branch"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def target_branch(self) -> str:
        result = self.info["target_branch"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def sha(self) -> str:
        result = self.info["sha"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def squash(self) -> bool:
        result = self.info.get(
            "squash", False
        )  # missing means auto-squash not supported
        if TYPE_CHECKING:
            assert isinstance(result, bool)
        return result

    @property
    def squash_on_merge(self) -> bool:
        result = self.info.get("squash_on_merge", False)
        if TYPE_CHECKING:
            assert isinstance(result, bool)
        return result

    @property
    def source_project_id(self) -> int:
        result = self.info["source_project_id"]
        if TYPE_CHECKING:
            assert isinstance(result, int)
        return result

    @property
    def target_project_id(self) -> int:
        result = self.info["target_project_id"]
        if TYPE_CHECKING:
            assert isinstance(result, int)
        return result

    @property
    def draft(self) -> bool:
        result = self.info["draft"]
        if TYPE_CHECKING:
            assert isinstance(result, bool)
        return result

    @property
    def web_url(self) -> str:
        result = self.info["web_url"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def blocking_discussions_resolved(self) -> bool:
        result = self.info["blocking_discussions_resolved"]
        if TYPE_CHECKING:
            assert isinstance(result, bool)
        return result

    @property
    def force_remove_source_branch(self) -> bool:
        result = self.info["force_remove_source_branch"]
        if TYPE_CHECKING:
            assert isinstance(result, bool)
        return result

    def update_sha(self, sha: str) -> None:
        """record the updated sha. We don't use refetch_info instead as it may hit cache."""
        self._info["sha"] = sha

    def refetch_info(self) -> None:
        result = self._api.call(
            gitlab.GET(
                f"/projects/{self.project_id}/merge_requests/{self.iid}",
                {"include_rebase_in_progress": "true"},
            )
        )
        if TYPE_CHECKING:
            assert isinstance(result, dict)
        self._info = result

    def comment(self, message: str) -> dict[str, Any]:
        notes_url = f"/projects/{self.project_id}/merge_requests/{self.iid}/notes"
        result = self._api.call(gitlab.POST(notes_url, {"body": message}))
        if TYPE_CHECKING:
            assert isinstance(result, dict)
        return result

    def rebase(self) -> None:
        self.refetch_info()

        if not self.rebase_in_progress:
            self._api.call(
                gitlab.PUT(
                    f"/projects/{self.project_id}/merge_requests/{self.iid}/rebase"
                )
            )
        else:
            # We wanted to rebase and someone just happened to press the button for us!
            log.info("A rebase was already in progress on the merge request!")

        max_attempts = 30
        wait_between_attempts_in_secs = 1

        for _ in range(max_attempts):
            self.refetch_info()
            if not self.rebase_in_progress:
                if self.merge_error:
                    raise MergeRequestRebaseFailed(self.merge_error)
                return

            time.sleep(wait_between_attempts_in_secs)

        raise TimeoutError("Waiting for merge request to be rebased by GitLab")

    def accept(
        self,
        remove_branch: bool = False,
        sha: Optional[str] = None,
        merge_when_pipeline_succeeds: bool = True,
        squash_on_merge: bool = False,
    ) -> dict[str, Any]:
        result = self._api.call(
            gitlab.PUT(
                f"/projects/{self.project_id}/merge_requests/{self.iid}/merge",
                {
                    "should_remove_source_branch": remove_branch,
                    "merge_when_pipeline_succeeds": merge_when_pipeline_succeeds,
                    "sha": sha
                    or self.sha,  # if provided, ensures what is merged is what we want (or fails)
                    "squash": squash_on_merge,
                },
            )
        )
        if TYPE_CHECKING:
            assert isinstance(result, dict)
        return result

    def close(self) -> dict[str, Any]:
        result = self._api.call(
            gitlab.PUT(
                f"/projects/{self.project_id}/merge_requests/{self.iid}",
                {"state_event": "close"},
            )
        )
        if TYPE_CHECKING:
            assert isinstance(result, dict)
        return result

    def assign_to(self, user_id: int) -> dict[str, Any]:
        result = self._api.call(
            gitlab.PUT(
                f"/projects/{self.project_id}/merge_requests/{self.iid}",
                {"assignee_id": user_id},
            )
        )
        if TYPE_CHECKING:
            assert isinstance(result, dict)
        return result

    def unassign(self) -> dict[str, Any]:
        return self.assign_to(0)

    def fetch_approvals(
        self, approvals_factory: Callable[[gitlab.Api, dict[str, Any]], Approvals]
    ) -> Approvals:
        info = {"iid": self.iid, "project_id": self.project_id}
        approvals = approvals_factory(self.api, info)
        approvals.refetch_info()
        return approvals

    def fetch_commits(self) -> list[dict[str, Any]]:
        result = self._api.call(
            gitlab.GET(f"/projects/{self.project_id}/merge_requests/{self.iid}/commits")
        )
        if TYPE_CHECKING:
            assert isinstance(result, list)
        return result

    def trigger_pipeline(self) -> dict[str, Any]:
        """
        Triggers a pipeline for the merge request.

        At first, try to trigger a merge request pipeline, which is different
        from a normal Gitlab CI pipeline and should be configured[0].
        If this fails due to unavailable merge request job definitions, trigger
        a normal pipeline for the source branch.

        [0]: https://docs.gitlab.com/ee/ci/pipelines/merge_request_pipelines.html
        """
        try:
            result = self._api.call(
                gitlab.POST(
                    f"/projects/{self.project_id}/merge_requests/{self.iid}/pipelines"
                )
            )
        except gitlab.BadRequest as exc:
            if exc.error_message is None:
                raise

            should_trigger_branch = False

            if NO_JOBS_MESSAGE in exc.error_message:
                # Failed due to unavailable merge request job definitions
                should_trigger_branch = True
                log.info(
                    "The pipeline is not configured for MR jobs, triggering a usual pipeline."
                )

            elif RULES_PREVENT_JOBS_MESSAGE in exc.error_message:
                # This does not necessarily always mean that we just need to trigger a branch pipeline,
                # but that is one possible reason. Worth trying.

                log.info(
                    "Error trying to trigger MR pipeline: '%s', triggering a branch pipeline as fallback.",
                    exc.error_message,
                )
                should_trigger_branch = True

            if not should_trigger_branch:
                # Propagate the error if we are not going to just trigger the branch pipeline
                raise

            result = self._api.call(
                gitlab.POST(
                    f"/projects/{self.project_id}/pipeline?ref={self.source_branch}"
                )
            )
        if TYPE_CHECKING:
            assert isinstance(result, dict)
        return result


class MergeRequestRebaseFailed(Exception):
    pass



import functools
import logging as log
import shlex
from collections import defaultdict
from datetime import timedelta

from marge import git


class RepoMock(git.Repo):
    @classmethod
    def init_for_merge_request(
        cls, merge_request, initial_target_sha, project, forked_project=None
    ):
        assert bool(forked_project) == (
            merge_request.source_project_id != merge_request.target_project_id
        )

        target_url = project.ssh_url_to_repo
        source_url = forked_project.ssh_url_to_repo if forked_project else target_url

        remote_repos = defaultdict(GitRepoModel)
        remote_repos[source_url].set_ref(merge_request.source_branch, merge_request.sha)
        remote_repos[target_url].set_ref(
            merge_request.target_branch, initial_target_sha
        )

        result = cls(
            remote_url=target_url,
            local_path="/tmp/blah",
            user_name="",
            user_email="",
            ssh_key_file="/home/homer/.ssh/id_rsa",
            timeout=timedelta(seconds=1000000),
            reference="the_reference",
            keep_committers=False,
        )

        # pylint: disable=attribute-defined-outside-init
        result.mock_impl = GitModel(origin=target_url, remote_repos=remote_repos)
        return result

    def git(self, *args, from_repo=True, keep_committers=False):
        command = args[0]
        command_args = args[1:]

        log.info("Run: git %r %s", command, " ".join(map(repr, command_args)))
        assert from_repo == (command != "clone")

        command_impl_name = command.replace("-", "_")
        command_impl = getattr(self.mock_impl, command_impl_name, None)
        assert command_impl, f"git: Unexpected command {command}"
        try:
            result = command_impl(*command_args)
        except Exception:
            log.warning("Failed to simulate: git %r %s", command, command_args)
            raise
        return self._pretend_result_comes_from_popen(result)

    @staticmethod
    def _pretend_result_comes_from_popen(result):
        result_bytes = ("" if result is None else str(result)).encode("ascii")
        return stub(stdout=result_bytes)


class stub:  # pylint: disable=invalid-name,too-few-public-methods
    def __init__(self, **kwargs):
        self.__dict__ = kwargs


class GitRepoModel:
    def __init__(self, copy_of=None):
        # pylint: disable=protected-access
        self._refs = dict(copy_of._refs) if copy_of else {}

    def set_ref(self, ref, commit):
        self._refs[ref] = commit

    def get_ref(self, ref):
        return self._refs[ref]

    def has_ref(self, ref):
        return ref in self._refs

    def del_ref(self, ref):
        self._refs.pop(ref, None)

    def __repr__(self):
        return f"<{type(self)}: {self._refs}>"


class GitModel:
    def __init__(self, origin, remote_repos):
        assert origin in remote_repos

        self.remote_repos = remote_repos
        self._local_repo = GitRepoModel()
        self._remotes = {"origin": origin}
        self._remote_refs = {}
        self._branch = None
        self.on_push_callbacks = []

    @property
    def _head(self):
        return self._local_repo.get_ref(self._branch)

    def remote(self, *args):
        action = args[0]
        if action == "rm":
            _, remote = args
            try:
                self._remotes.pop(remote)
            except KeyError as err:
                raise git.GitError(f"No such remote: {remote}") from err

        elif action == "add":
            _, remote, url = args
            self._remotes[remote] = url
        else:
            assert False, args

    def fetch(self, *args):
        _, remote_name = args
        assert args == ("--prune", remote_name)
        remote_url = self._remotes[remote_name]
        remote_repo = self.remote_repos[remote_url]
        self._remote_refs[remote_name] = GitRepoModel(copy_of=remote_repo)

    def checkout(self, *args):
        if args[0] == "-B":  # -B == create if it doesn't exist
            _, branch, start_point, _ = args
            assert args == ("-B", branch, start_point, "--")
            assert start_point == "" or "/" in start_point  # '' when "local"

            # create if it doesn't exist
            if not self._local_repo.has_ref(branch):
                if start_point:
                    remote_name, remote_branch = start_point.split("/")
                    assert remote_branch == branch

                    remote_url = self._remotes[remote_name]
                    remote_repo = self.remote_repos[remote_url]
                    commit = remote_repo.get_ref(branch)
                    self._local_repo.set_ref(branch, commit)
                else:
                    self._local_repo.set_ref(branch, self._head)
        else:
            branch, _ = args
            assert args == (branch, "--")
            assert self._local_repo.has_ref(branch)

        # checkout
        self._branch = branch

    def branch(self, *args):
        if args[0] == "-D":
            _, branch = args
            assert self._branch != branch
            self._local_repo.del_ref(branch)
        else:
            assert False

    def rev_parse(self, arg):
        if arg == "HEAD":
            return self._head

        remote, branch = arg.split("/")
        return self._remote_refs[remote].get_ref(branch)

    def rebase(self, arg):
        remote, branch = arg.split("/")
        new_base = self._remote_refs[remote].get_ref(branch)
        if new_base != self._head:
            new_sha = f"rebase({self._head} onto {new_base})"
            self._local_repo.set_ref(self._branch, new_sha)

    def merge(self, arg):
        remote, branch = arg.split("/")

        other_ref = self._remote_refs[remote].get_ref(branch)
        if other_ref != self._head:
            new_sha = f"merge({self._head} with {other_ref})"
            self._local_repo.set_ref(self._branch, new_sha)

    def push(self, *args):
        force_flag, remote_name, refspec = args

        assert force_flag in ("", "--force-with-lease")

        branch, remote_branch = refspec.split(":")
        remote_url = self._remotes[remote_name]
        remote_repo = self.remote_repos[remote_url]

        old_sha = remote_repo.get_ref(remote_branch)
        new_sha = self._local_repo.get_ref(branch)

        if force_flag:
            remote_repo.set_ref(remote_branch, new_sha)
        else:
            expected_remote_sha = self._remote_refs[remote_name].get_ref(remote_branch)
            if old_sha != expected_remote_sha:
                raise git.GitError("conflict: can't push")
            remote_repo.set_ref(remote_branch, new_sha)

        for callback in self.on_push_callbacks:
            callback(
                remote_url=remote_url,
                remote_branch=remote_branch,
                old_sha=old_sha,
                new_sha=new_sha,
            )

    def config(self, *args):
        if len(args) == 2 and args[0] == "--get":
            _, remote, _ = elems = args[1].split(".")
            assert elems == ["remote", remote, "url"], elems
            return self._remotes[remote]
        return ""

    def diff_index(self, *args):
        assert args == ("--quiet", "HEAD")
        # we don't model dirty index

    def ls_files(self, *args):
        assert args == ("--others",)
        # we don't model untracked files

    def filter_branch(self, *args):
        _, _, filter_cmd, commit_range = args
        assert args == ("--force", "--msg-filter", filter_cmd, commit_range)

        trailers_var, keep_trailers_var, python, script_path = shlex.split(filter_cmd)
        _, trailers_str = trailers_var.split("=")
        _, keep_trailers_str = keep_trailers_var.split("=")

        assert trailers_var == f"TRAILERS={trailers_str}"
        assert keep_trailers_var == f"KEEP_TRAILERS={keep_trailers_str}"
        assert python == "python3"
        assert script_path.endswith("marge/trailerfilter.py")

        trailers = list(
            sorted(set(line.split(":")[0] for line in trailers_str.split("\n")))
        )
        assert trailers

        new_sha = functools.reduce(
            lambda x, f: f"add-{f}({x})",
            [trailer.lower() for trailer in trailers],
            self._head,
        )
        self._local_repo.set_ref(self._branch, new_sha)
        return new_sha



import contextlib
import dataclasses
import datetime
import logging as log
import os
import re
import shlex
import subprocess
import sys
from typing import Any, Iterator, Optional

from . import keepcommitters, trailerfilter

# Turning off StrictHostKeyChecking is a nasty hack to approximate
# just accepting the hostkey sight unseen the first time marge
# connects. The proper solution would be to pass in known_hosts as
# a commandline parameter, but in practice few people will bother anyway and
# in this case the threat of MiTM seems somewhat bogus.
GIT_SSH_COMMAND = "ssh -o StrictHostKeyChecking=no "


def _filter_branch_script(
    trailer_name: str, trailer_values: list[str], keep_trailers: bool
) -> str:
    trailers: str = shlex.quote(
        "\n".join(
            f"{trailer_name}: {trailer_value}"
            for trailer_value in trailer_values or [""]
        )
    )

    filter_script = f"TRAILERS={trailers} KEEP_TRAILERS={keep_trailers} python3 {trailerfilter.__file__}"
    return filter_script


@dataclasses.dataclass
class Repo:
    remote_url: str
    local_path: str
    user_email: str
    user_name: str
    ssh_key_file: Optional[str]
    timeout: Optional[datetime.timedelta]
    reference: Optional[str]
    keep_committers: bool
    auth_token: dataclasses.InitVar[Optional[str]] = None

    def __post_init__(self, auth_token: Optional[str] = None) -> None:
        if auth_token:
            credentials = "oauth2:" + auth_token
            # insert token auth "oauth2:<auth_token>@"
            pattern = "(http(s)?://)"
            replacement = r"\1" + credentials + "@"
            self.git_url = re.sub(pattern, replacement, self.remote_url, 1)
        else:
            self.git_url = self.remote_url

    def clone(self) -> None:
        reference_flag = "--reference=" + self.reference if self.reference else ""
        self.git(
            "clone",
            "--origin=origin",
            "--filter=blob:none",
            reference_flag,
            self.git_url,
            self.local_path,
            from_repo=False,
        )

    def configure(self) -> None:
        self.git("config", "user.email", self.user_email)
        self.git("config", "user.name", self.user_name)
        if self.ssh_key_file:
            self.git("config", "gpg.format", "ssh")
            self.git("config", "user.signingKey", self.ssh_key_file)

    def configure_remote(self, remote_name: str, remote_url: str) -> None:
        try:
            self.git("remote", "rm", remote_name)
        except GitError:
            pass
        self.git("remote", "add", remote_name, remote_url)
        self.git("config", f"remote.{remote_name}.partialclonefilter", "blob:none")

    def fetch(self, remote_name: str, remote_url: Optional[str] = None) -> None:
        if remote_name != "origin":
            assert remote_url is not None
            self.configure_remote(remote_name, remote_url)
        self.git("fetch", "--prune", remote_name)

    @contextlib.contextmanager
    def _handle_filter_branch(self, branch: str) -> Iterator[None]:
        try:
            yield
        except GitError:
            log.warning("filter-branch failed, will try to restore")
            try:
                self.get_commit_hash("refs/original/refs/heads/")
            except GitError:
                log.warning("No changes have been effected by filter-branch")
            else:
                self.git("reset", "--hard", "refs/original/refs/heads/" + branch)
            raise

    def sign_commits(self, branch: str, start_commit: str) -> str:
        """Sign commits in `branch` from `start_commit`"""
        commit_range = start_commit + ".." + branch
        with self._handle_filter_branch(branch):
            self.git(
                "filter-branch",
                "--force",
                "--commit-filter",
                "git commit-tree -S $@",
                "--env-filter",
                f'GIT_COMMITTER_EMAIL="{self.user_email}" GIT_COMMITTER_NAME="{self.user_name}"',
                commit_range,
            )
        return self.get_commit_hash()

    def tag_with_trailer(
        self,
        trailer_name: str,
        trailer_values: list[str],
        branch: str,
        start_commit: str,
        keep_trailers: bool,
    ) -> str:
        """Add `trailer_name` in commit messages with `trailer_values` in `branch` from `start_commit`."""

        # Strips all `$trailer_name``: lines and trailing newlines, adds an empty
        # newline and tags on the `$trailer_name: $trailer_value` for each `trailer_value` in
        # `trailer_values`.
        filter_script = _filter_branch_script(
            trailer_name, trailer_values, keep_trailers
        )
        commit_range = start_commit + ".." + branch
        with self._handle_filter_branch(branch):
            # --force = overwrite backup of last filter-branch
            self.git(
                "filter-branch", "--force", "--msg-filter", filter_script, commit_range
            )
        return self.get_commit_hash()

    def merge(
        self,
        source_branch: str,
        target_branch: str,
        *merge_args: str,
        source_repo_url: Optional[str] = None,
        local: bool = False,
    ) -> str:
        """Merge `target_branch` into `source_branch` and return the new HEAD commit id.

        By default `source_branch` and `target_branch` are assumed to reside in the same
        repo as `self`. However, if `source_repo_url` is passed and not `None`,
        `source_branch` is taken from there.

        Throws a `GitError` if the merge fails. Will also try to --abort it.
        """
        return self._fuse_branch(
            "merge",
            source_branch,
            target_branch,
            *merge_args,
            source_repo_url=source_repo_url,
            local=local,
        )

    def fast_forward(
        self,
        source: str,
        target: str,
        source_repo_url: Optional[str] = None,
        local: bool = False,
    ) -> str:
        return self.merge(
            source,
            target,
            "--ff",
            "--ff-only",
            source_repo_url=source_repo_url,
            local=local,
        )

    def rebase(
        self,
        branch: str,
        new_base: str,
        source_repo_url: Optional[str] = None,
        local: bool = False,
    ) -> str:
        """Rebase `new_base` into `branch` and return the new HEAD commit id.

        By default `branch` and `new_base` are assumed to reside in the same
        repo as `self`. However, if `source_repo_url` is passed and not `None`,
        `branch` is taken from there.

        Throws a `GitError` if the rebase fails. Will also try to --abort it.
        """
        return self._fuse_branch(
            "rebase",
            branch,
            new_base,
            source_repo_url=source_repo_url,
            local=local,
            keep_committers=self.keep_committers,
        )

    def _fuse_branch(
        self,
        strategy: str,
        branch: str,
        target_branch: str,
        *fuse_args: Any,
        source_repo_url: Optional[str] = None,
        local: bool = False,
        keep_committers: bool = False,
    ) -> str:
        assert source_repo_url or branch != target_branch, branch

        if not local:
            self.fetch("origin")
            target = "origin/" + target_branch
            if source_repo_url:
                self.fetch("source", source_repo_url)
                self.checkout_branch(branch, "source/" + branch)
            else:
                self.checkout_branch(branch, "origin/" + branch)
        else:
            self.checkout_branch(branch)
            target = target_branch

        try:
            self.git(strategy, target, keep_committers=keep_committers, *fuse_args)
        except GitError:
            log.warning("%s failed, doing an --abort", strategy)
            self.git(strategy, "--abort")
            raise
        return self.get_commit_hash()

    def remove_branch(self, branch: str, *, new_current_branch: str = "master") -> None:
        assert branch != new_current_branch
        self.git("branch", "-D", branch)

    def checkout_branch(self, branch: str, start_point: str = "") -> None:
        create_and_reset = "-B" if start_point else ""
        self.git("checkout", create_and_reset, branch, start_point, "--")

    def push(
        self,
        branch: str,
        *,
        source_repo_url: Optional[str] = None,
        force: bool = False,
        skip_ci: bool = False,
    ) -> None:
        self.git("checkout", branch, "--")

        self.git("diff-index", "--quiet", "HEAD")  # check it is not dirty

        untracked_files = self.git(
            "ls-files", "--others"
        ).stdout  # check no untracked files
        if untracked_files:
            raise GitError("There are untracked files", untracked_files)

        if source_repo_url:
            assert self.get_remote_url("source") == source_repo_url
            source = "source"
        else:
            source = "origin"
        force_flag = "--force-with-lease" if force else ""
        skip_flag = ("-o", "ci.skip") if skip_ci else ()
        self.git("push", force_flag, *skip_flag, source, f"{branch}:{branch}")

    def get_commit_hash(self, rev: str = "HEAD") -> str:
        """Return commit hash for `rev` (default "HEAD")."""
        result = self.git("rev-parse", rev)
        return result.stdout.decode("ascii").strip()

    def get_remote_url(self, name: str) -> str:
        return (
            self.git("config", "--get", f"remote.{name}.url")
            .stdout.decode("utf-8")
            .strip()
        )

    def git(
        self, *args: str, from_repo: bool = True, keep_committers: bool = False
    ) -> "subprocess.CompletedProcess[bytes]":
        env = os.environ.copy()
        env["FILTER_BRANCH_SQUELCH_WARNING"] = "1"
        if self.ssh_key_file:
            # ssh's handling of identity files is infuriatingly dumb, to get it
            # to actually really use the IdentityFile we pass in via -i we also
            # need to tell it to ignore ssh-agent (IdentitiesOnly=true) and not
            # read in any identities from ~/.ssh/config etc (-F /dev/null),
            # because they append and it tries them in order, starting with config file
            env["GIT_SSH_COMMAND"] = " ".join(
                [
                    GIT_SSH_COMMAND,
                    "-F",
                    "/dev/null",
                    "-o",
                    "IdentitiesOnly=yes",
                    "-i",
                    self.ssh_key_file,
                ]
            )

        command = ["git"]
        if keep_committers:
            command.extend(
                (
                    "-c",
                    f"rebase.instructionFormat=%s%nexec python3 {keepcommitters.__file__} %H",
                )
            )

        if from_repo:
            command.extend(["-C", self.local_path])
        command.extend([arg for arg in args if str(arg)])

        log.debug("Running %s", " ".join(shlex.quote(w) for w in command))
        try:
            timeout_seconds = (
                self.timeout.total_seconds() if self.timeout is not None else None
            )
            return _run(*command, env=env, check=True, timeout=timeout_seconds)
        except subprocess.CalledProcessError as err:
            log.warning("git returned %s", err.returncode)
            log.warning("stdout: %r", err.stdout)
            log.warning("stderr: %r", err.stderr)
            raise GitError(err) from err


def _run(
    *args: Any,
    env: Optional[dict[str, str]] = None,
    check: bool = False,
    timeout: Optional[float] = None,
) -> "subprocess.CompletedProcess[bytes]":
    encoded_args = (
        [a.encode("utf-8") for a in args] if sys.platform != "win32" else args
    )
    return subprocess.run(
        encoded_args, capture_output=True, check=check, env=env, timeout=timeout
    )


class GitError(Exception):
    pass



# pylint: disable=too-many-lines,too-many-locals
import contextlib
import dataclasses
import functools
import subprocess
from datetime import timedelta
from unittest.mock import ANY, patch

import pytest

import marge.commit
import marge.git
import marge.gitlab
import marge.interval
import marge.job
import marge.project
import marge.single_merge_job
import marge.user
from marge.gitlab import GET, POST, PUT
from marge.job import Fusion
from marge.merge_request import MergeRequest
from tests import test_commit
from tests.git_repo_mock import RepoMock
from tests.gitlab_api_mock import Api, Error, MockedEndpointNotFound, MockLab, Ok
from tests.test_project import INFO as TEST_PROJECT_INFO

INITIAL_MR_SHA = test_commit.INFO["id"]


def _commit(commit_id, status):
    return {
        "id": commit_id,
        "short_id": commit_id,
        "author_name": "J. Bond",
        "author_email": "jbond@mi6.gov.uk",
        "message": "Shaken, not stirred",
        "status": status,
    }


def _branch(name, protected=False):
    return {"name": name, "protected": protected}


def _pipeline(sha1, status, ref="useless_new_feature"):
    return {
        "id": 47,
        "status": status,
        "ref": ref,
        "sha": sha1,
        "web_url": "https://link/pipelines/87",
    }


class SingleJobMockLab(MockLab):
    def __init__(
        self,
        *,
        initial_master_sha,
        rewritten_sha,
        gitlab_url=None,
        fork=False,
        expect_gitlab_rebase=False,
        merge_request_options=None,
        guarantee_final_pipeline=False,
    ):
        super().__init__(
            initial_master_sha,
            gitlab_url,
            fork=fork,
            merge_request_options=merge_request_options,
        )
        api = self.api
        self.rewritten_sha = rewritten_sha
        target_project_id = self.merge_request_info["project_id"]
        source_project_id = self.merge_request_info["source_project_id"]
        if expect_gitlab_rebase:
            api.add_transition(
                PUT(
                    f"/projects/{target_project_id}/merge_requests/"
                    f'{self.merge_request_info["iid"]}/rebase'
                ),
                Ok(True),
                from_state="initial",
                to_state="rebase-in-progress",
            )
            api.add_merge_request(
                dict(self.merge_request_info, rebase_in_progress=True),
                from_state="rebase-in-progress",
                to_state="rebase-finished",
            )
            api.add_merge_request(
                dict(
                    self.merge_request_info, rebase_in_progress=False, sha=rewritten_sha
                ),
                from_state="rebase-finished",
                to_state="pushed",
            )

        if guarantee_final_pipeline:
            # Corresponds to the `merge_request.trigger_pipeline()` call.
            api.add_transition(
                POST(
                    f"/projects/{target_project_id}/merge_requests/{self.merge_request_info['iid']}/pipelines"
                ),
                Ok({}),
                to_state="final_pipeline_triggered",
            )
            # Corresponds to `pipelines_by_merge_request()` by `wait_for_ci_to_pass`.
            api.add_pipelines(
                target_project_id,
                self.merge_request_info["iid"],
                _pipeline(
                    sha1=rewritten_sha,
                    status="success",
                    ref=self.merge_request_info["source_branch"],
                ),
                from_state=["final_pipeline_triggered"],
                to_state="pushed",
            )

        api.add_pipelines(
            target_project_id,
            self.merge_request_info["iid"],
            _pipeline(
                sha1=rewritten_sha,
                status="running",
                ref=self.merge_request_info["source_branch"],
            ),
            from_state="pushed",
            to_state="passed",
        )
        api.add_pipelines(
            target_project_id,
            self.merge_request_info["iid"],
            _pipeline(
                sha1=rewritten_sha,
                status="success",
                ref=self.merge_request_info["source_branch"],
            ),
            from_state=["passed", "merged"],
        )
        api.add_transition(
            GET(
                f"/projects/{source_project_id}/repository/branches/"
                f'{self.merge_request_info["source_branch"]}'
            ),
            Ok({"commit": _commit(commit_id=rewritten_sha, status="running")}),
            from_state="pushed",
        )
        api.add_transition(
            GET(
                f"/projects/{source_project_id}/repository/branches/"
                f'{self.merge_request_info["source_branch"]}'
            ),
            Ok({"commit": _commit(commit_id=rewritten_sha, status="success")}),
            from_state="passed",
        )
        api.add_transition(
            PUT(
                f'/projects/{target_project_id}/merge_requests/{self.merge_request_info["iid"]}/merge',
                {
                    "sha": rewritten_sha,
                    "should_remove_source_branch": True,
                    "merge_when_pipeline_succeeds": True,
                    "squash": False,
                },
            ),
            Ok({}),
            from_state=["passed", "skipped"],
            to_state="merged",
        )
        api.add_transition(
            PUT(
                f'/projects/{target_project_id}/merge_requests/{self.merge_request_info["iid"]}/merge',
                {
                    "sha": rewritten_sha,
                    "should_remove_source_branch": True,
                    "merge_when_pipeline_succeeds": True,
                    "squash": True,
                },
            ),
            Ok({}),
            from_state=["passed", "skipped"],
            to_state="merged",
        )
        api.add_merge_request(
            dict(self.merge_request_info, state="merged"), from_state="merged"
        )
        api.add_transition(
            GET(
                f"/projects/{target_project_id}/repository/branches/"
                f'{self.merge_request_info["target_branch"]}'
            ),
            Ok({"commit": {"id": self.rewritten_sha}}),
            from_state="merged",
        )
        api.expected_note(
            self.merge_request_info,
            "A merge request may have skipped the queue, trying again...",
            from_state=["pushed_but_master_moved", "merge_rejected"],
        )
        api.expected_note(
            self.merge_request_info,
            "Unexpected exception in bot while handling this MR.",
        )

    def push_updated(self, remote_url, remote_branch, old_sha, new_sha):
        source_project = self.forked_project_info or self.project_info
        assert remote_url == source_project["ssh_url_to_repo"]
        assert remote_branch == self.merge_request_info["source_branch"]
        assert old_sha == INITIAL_MR_SHA
        assert new_sha == self.rewritten_sha
        self.api.state = "pushed"

    @contextlib.contextmanager
    def expected_failure(self, message):
        author_assigned = False

        def assign_to_author():
            nonlocal author_assigned
            author_assigned = True

        self.api.add_transition(
            PUT(
                f'/projects/1234/merge_requests/{self.merge_request_info["iid"]}',
                args={"assignee_id": self.author_id},
            ),
            assign_to_author,
        )
        error_note = f"This branch couldn't be merged: {message}"
        self.api.expected_note(self.merge_request_info, error_note)

        yield

        assert author_assigned
        assert error_note in self.api.notes


class TestUpdateAndAccept:  # pylint: disable=too-many-public-methods
    @dataclasses.dataclass
    class Mocks:
        mocklab: SingleJobMockLab
        api: Api
        job: marge.single_merge_job.SingleMergeJob

    @pytest.fixture(params=[True, False])
    def fork(self, request):
        return request.param

    @pytest.fixture(params=list(Fusion))
    def fusion(self, request):
        return request.param

    @pytest.fixture(params=[True, False])
    def add_tested(self, request):
        return request.param

    @pytest.fixture(params=[True, False])
    def add_part_of(self, request):
        return request.param

    @pytest.fixture(params=[False])  # TODO: Needs support in mocklab
    def add_reviewers(self, request):
        return request.param

    @pytest.fixture(params=[True, False])
    def guarantee_final_pipeline(self, request):
        return request.param

    @pytest.fixture()
    def options_factory(
        self, fusion, add_tested, add_reviewers, add_part_of, guarantee_final_pipeline
    ):
        def make_options(**kwargs):
            fixture_opts = {
                "fusion": fusion,
                "add_tested": add_tested,
                "add_part_of": add_part_of,
                "add_reviewers": add_reviewers,
                "guarantee_final_pipeline": guarantee_final_pipeline,
            }
            assert not set(fixture_opts).intersection(kwargs)
            kwargs.update(fixture_opts)
            return marge.job.MergeJobOptions(**kwargs)

        yield make_options

    @pytest.fixture()
    def update_sha(self, fusion):
        def new_sha(new, old):
            pats = {
                marge.job.Fusion.rebase: "rebase(%s onto %s)",
                marge.job.Fusion.merge: "merge(%s with %s)",
                marge.job.Fusion.gitlab_rebase: "rebase(%s onto %s)",
            }
            return pats[fusion] % (new, old)

        yield new_sha

    @pytest.fixture()
    def rewrite_sha(self, fusion, add_tested, add_reviewers, add_part_of):
        def new_sha(sha):
            # NB. The order matches the one used in the Git mock to run filters
            if add_tested and fusion == marge.job.Fusion.rebase:
                sha = f"add-tested-by({sha})"

            if add_reviewers and fusion != marge.job.Fusion.gitlab_rebase:
                sha = f"add-reviewed-by({sha})"

            if add_part_of and fusion != marge.job.Fusion.gitlab_rebase:
                sha = f"add-part-of({sha})"

            return sha

        yield new_sha

    @pytest.fixture(autouse=True)
    def patch_sleep(self):
        with patch("time.sleep"):
            yield

    @pytest.fixture()
    def mocklab_factory(self, fork, fusion, guarantee_final_pipeline):
        expect_rebase = fusion is Fusion.gitlab_rebase
        return functools.partial(
            SingleJobMockLab,
            fork=fork,
            expect_gitlab_rebase=expect_rebase,
            guarantee_final_pipeline=guarantee_final_pipeline,
        )

    @pytest.fixture()
    def mocks_factory(self, mocklab_factory, options_factory, update_sha, rewrite_sha):
        # pylint: disable=too-many-locals
        def make_mocks(
            initial_master_sha=None,
            rewritten_sha=None,
            extra_opts=None,
            extra_mocklab_opts=None,
            on_push=None,
        ):
            options = options_factory(**(extra_opts or {}))
            initial_master_sha = initial_master_sha or "505050505e"

            if not rewritten_sha:
                rewritten_sha = rewrite_sha(
                    update_sha(INITIAL_MR_SHA, initial_master_sha)
                )

            mocklab = mocklab_factory(
                initial_master_sha=initial_master_sha,
                rewritten_sha=rewritten_sha,
                **(extra_mocklab_opts or {}),
            )
            api = mocklab.api

            project_id = mocklab.project_info["id"]
            merge_request_iid = mocklab.merge_request_info["iid"]

            project = marge.project.Project.fetch_by_id(project_id, api)
            forked_project = None
            if mocklab.forked_project_info:
                forked_project_id = mocklab.forked_project_info["id"]
                forked_project = marge.project.Project.fetch_by_id(
                    forked_project_id, api
                )

            merge_request = MergeRequest.fetch_by_iid(
                project_id, merge_request_iid, api
            )

            def assert_can_push(*_args, **_kwargs):
                assert options.fusion is not Fusion.gitlab_rebase

            callback = on_push or mocklab.push_updated
            repo = RepoMock.init_for_merge_request(
                merge_request=merge_request,
                initial_target_sha=mocklab.initial_master_sha,
                project=project,
                forked_project=forked_project,
            )
            repo.mock_impl.on_push_callbacks.append(assert_can_push)
            repo.mock_impl.on_push_callbacks.append(callback)

            user = marge.user.User.myself(api)
            job = marge.single_merge_job.SingleMergeJob(
                api=api,
                user=user,
                project=project,
                merge_request=merge_request,
                repo=repo,
                options=options,
            )
            return self.Mocks(mocklab=mocklab, api=api, job=job)

        yield make_mocks

    @pytest.fixture()
    def mocks(self, mocks_factory):
        yield mocks_factory()

    def test_succeeds_first_time(self, mocks):
        mocks.job.execute()
        assert mocks.api.state == "merged"
        assert mocks.api.notes == []

    def test_succeeds_with_updated_branch(self, mocks):
        mocks.api.add_transition(
            GET(
                f"/projects/1234/repository/branches/"
                f'{mocks.mocklab.merge_request_info["source_branch"]}'
            ),
            Ok({"commit": {"id": mocks.mocklab.rewritten_sha}}),
            from_state="initial",
            to_state="pushed",
        )
        mocks.job.execute()

        assert mocks.api.state == "merged"
        assert mocks.api.notes == []

    def test_succeeds_if_skipped(self, mocks):
        mocks.api.add_pipelines(
            mocks.mocklab.merge_request_info["target_project_id"],
            mocks.mocklab.merge_request_info["iid"],
            _pipeline(sha1=mocks.mocklab.rewritten_sha, status="running"),
            from_state="pushed",
            to_state="skipped",
        )
        mocks.api.add_pipelines(
            mocks.mocklab.merge_request_info["target_project_id"],
            mocks.mocklab.merge_request_info["iid"],
            _pipeline(sha1=mocks.mocklab.rewritten_sha, status="skipped"),
            from_state=["skipped", "merged"],
        )
        # `pipelines_by_merge_request()` by `wait_for_ci_to_pass`.
        mocks.api.add_pipelines(
            mocks.mocklab.merge_request_info["target_project_id"],
            mocks.mocklab.merge_request_info["iid"],
            _pipeline(
                sha1=mocks.mocklab.rewritten_sha,
                status="skipped",
                ref=mocks.mocklab.merge_request_info["source_branch"],
            ),
            from_state=["final_pipeline_triggered"],
            to_state="passed",
        )
        mocks.job.execute()

        assert mocks.api.state == "merged"
        assert mocks.api.notes == []

    def test_succeeds_if_source_is_master(self, mocks_factory):
        mocks = mocks_factory(
            extra_mocklab_opts={
                "merge_request_options": {
                    "source_branch": "master",
                    "target_branch": "production",
                }
            }
        )
        mocks.api.add_transition(
            GET(
                f"/projects/1234/repository/branches/"
                f'{mocks.mocklab.merge_request_info["source_branch"]}'
            ),
            Ok({"commit": {"id": mocks.mocklab.rewritten_sha}}),
            from_state="initial",
            to_state="pushed",
        )
        mocks.job.execute()

        assert mocks.api.state == "merged"
        assert mocks.api.notes == []

    def test_fails_if_ci_fails(self, mocks):
        mocks.api.add_pipelines(
            mocks.mocklab.merge_request_info["target_project_id"],
            mocks.mocklab.merge_request_info["iid"],
            _pipeline(sha1=mocks.mocklab.rewritten_sha, status="running"),
            from_state="pushed",
            to_state="failed",
        )
        mocks.api.add_pipelines(
            mocks.mocklab.merge_request_info["target_project_id"],
            mocks.mocklab.merge_request_info["iid"],
            _pipeline(sha1=mocks.mocklab.rewritten_sha, status="failed"),
            from_state=["failed"],
        )
        mocks.api.add_pipelines(
            mocks.mocklab.merge_request_info["target_project_id"],
            mocks.mocklab.merge_request_info["iid"],
            _pipeline(
                sha1=mocks.mocklab.rewritten_sha,
                status="failed",
                ref=mocks.mocklab.merge_request_info["source_branch"],
            ),
            from_state=["final_pipeline_triggered"],
            to_state="failed",
        )

        expected_message = "CI failed! See pipeline https://link/pipelines/87."
        with mocks.mocklab.expected_failure(expected_message):
            mocks.job.execute()

        assert mocks.api.state == "failed"

    def test_fails_if_ci_canceled(self, mocks):
        mocks.api.add_pipelines(
            mocks.mocklab.merge_request_info["target_project_id"],
            mocks.mocklab.merge_request_info["iid"],
            _pipeline(sha1=mocks.mocklab.rewritten_sha, status="running"),
            from_state="pushed",
            to_state="canceled",
        )
        mocks.api.add_pipelines(
            mocks.mocklab.merge_request_info["target_project_id"],
            mocks.mocklab.merge_request_info["iid"],
            _pipeline(sha1=mocks.mocklab.rewritten_sha, status="canceled"),
            from_state=["canceled"],
        )
        mocks.api.add_pipelines(
            mocks.mocklab.merge_request_info["target_project_id"],
            mocks.mocklab.merge_request_info["iid"],
            _pipeline(
                sha1=mocks.mocklab.rewritten_sha,
                status="canceled",
                ref=mocks.mocklab.merge_request_info["source_branch"],
            ),
            from_state=["final_pipeline_triggered"],
            to_state="canceled",
        )

        expected_message = (
            "The CI run was canceled. See pipeline https://link/pipelines/87."
        )
        with mocks.mocklab.expected_failure(expected_message):
            mocks.job.execute()

        assert mocks.api.state == "canceled"

    def test_fails_on_not_acceptable_if_master_did_not_move(self, mocks):
        new_branch_head_sha = "99ba110035"
        mocks.api.add_transition(
            GET(
                f'/projects/{mocks.mocklab.merge_request_info["source_project_id"]}/'
                f"repository/branches/useless_new_feature"
            ),
            Ok({"commit": _commit(commit_id=new_branch_head_sha, status="success")}),
            from_state="pushed",
            to_state="pushed_but_head_changed",
        )
        mocks.api.add_pipelines(
            mocks.mocklab.merge_request_info["target_project_id"],
            mocks.mocklab.merge_request_info["iid"],
            _pipeline(
                sha1=new_branch_head_sha,
                status="success",
                ref=mocks.mocklab.merge_request_info["source_branch"],
            ),
            from_state=["pushed_but_head_changed"],
        )

        with mocks.mocklab.expected_failure(
            "Someone pushed to branch while we were trying to merge"
        ):
            mocks.job.execute()

        assert mocks.api.state == "pushed_but_head_changed"
        assert mocks.api.notes == [
            "This branch couldn't be merged: Someone pushed to branch while we were trying to merge"
        ]

    def test_fails_if_branch_is_protected(self, mocks_factory, fusion):
        def reject_push(*_args, **_kwargs):
            raise marge.git.GitError()

        mocks = mocks_factory(on_push=reject_push)
        mocks.api.add_transition(
            GET(
                f'/projects/{mocks.mocklab.merge_request_info["source_project_id"]}/'
                f"repository/branches/useless_new_feature"
            ),
            Ok(_branch("useless_new_feature", protected=True)),
            from_state="initial",
            to_state="protected",
        )

        if fusion is Fusion.gitlab_rebase:
            mocks.api.add_transition(
                PUT(
                    f'/projects/{mocks.mocklab.merge_request_info["project_id"]}/'
                    f'merge_requests/{mocks.mocklab.merge_request_info["iid"]}/rebase'
                ),
                Error(
                    marge.gitlab.MethodNotAllowed(
                        405, {"message": "405 Method Not Allowed"}
                    )
                ),
                from_state="initial",
            )

        with mocks.mocklab.expected_failure(
            "The branch is protected and therefore cannot be rebased. "
            "Either relax the branch protection rules or opt for a different merge strategy."
        ):
            mocks.job.execute()

        assert mocks.api.state == "protected"

    def test_second_time_if_master_moved(
        self, mocks_factory, fusion, update_sha, rewrite_sha
    ):
        initial_master_sha = "eaeaea9e9e"
        moved_master_sha = "fafafa"
        first_rewritten_sha = rewrite_sha(
            update_sha(INITIAL_MR_SHA, initial_master_sha)
        )
        second_rewritten_sha = rewrite_sha(
            update_sha(first_rewritten_sha, moved_master_sha)
        )

        # pylint: disable=unused-argument
        def push_effects(remote_url, remote_branch, old_sha, new_sha):
            if mocks.api.state == "initial":
                assert old_sha == INITIAL_MR_SHA
                assert new_sha == first_rewritten_sha
                mocks.api.state = "pushed_but_master_moved"
                remote_target_repo.set_ref(target_branch, moved_master_sha)
            elif mocks.api.state == "merge_rejected":
                assert new_sha == second_rewritten_sha
                mocks.api.state = "pushed"

        mocks = mocks_factory(
            initial_master_sha=initial_master_sha,
            rewritten_sha=second_rewritten_sha,
            on_push=push_effects,
        )

        source_project_info = (
            mocks.mocklab.forked_project_info or mocks.mocklab.project_info
        )
        target_project_info = mocks.mocklab.project_info

        source_project_url = source_project_info["ssh_url_to_repo"]
        target_project_url = target_project_info["ssh_url_to_repo"]

        source_branch = mocks.mocklab.merge_request_info["source_branch"]
        target_branch = mocks.mocklab.merge_request_info["target_branch"]

        remote_source_repo = mocks.job.repo.mock_impl.remote_repos[source_project_url]
        remote_target_repo = mocks.job.repo.mock_impl.remote_repos[target_project_url]

        mocks.api.add_merge_request(
            dict(mocks.mocklab.merge_request_info, sha=first_rewritten_sha),
            from_state=["pushed_but_master_moved", "merge_rejected"],
        )
        mocks.api.add_pipelines(
            mocks.mocklab.merge_request_info["target_project_id"],
            mocks.mocklab.merge_request_info["iid"],
            _pipeline(sha1=first_rewritten_sha, status="success"),
            from_state=["pushed_but_master_moved", "merge_rejected"],
        )
        mocks.api.add_transition(
            PUT(
                f"/projects/1234/merge_requests/"
                f'{mocks.mocklab.merge_request_info["iid"]}/merge',
                {
                    "sha": first_rewritten_sha,
                    "should_remove_source_branch": True,
                    "merge_when_pipeline_succeeds": True,
                    "squash": False,
                },
            ),
            Error(marge.gitlab.Unprocessable()),
            from_state="pushed_but_master_moved",
            to_state="merge_rejected",
        )
        mocks.api.add_transition(
            GET(
                f'/projects/{mocks.mocklab.merge_request_info["source_project_id"]}/'
                f"repository/branches/useless_new_feature"
            ),
            Ok({"commit": _commit(commit_id=first_rewritten_sha, status="success")}),
            from_state="pushed_but_master_moved",
        )
        mocks.api.add_transition(
            GET("/projects/1234/repository/branches/master"),
            Ok({"commit": _commit(commit_id=moved_master_sha, status="success")}),
            from_state="merge_rejected",
        )
        # Overwrite original `guarantee_final_pipeline` transition: no need in
        # the state changing here.
        mocks.api.add_transition(
            POST(
                f"/projects/1234/merge_requests/{mocks.mocklab.merge_request_info['iid']}/pipelines"
            ),
            Ok({}),
        )
        # Register additional pipeline check introduced by
        # `guarantee_final_pipeline`.
        mocks.api.add_pipelines(
            mocks.mocklab.merge_request_info["target_project_id"],
            mocks.mocklab.merge_request_info["iid"],
            _pipeline(
                sha1=first_rewritten_sha,
                status="success",
                ref=mocks.mocklab.merge_request_info["source_branch"],
            ),
            from_state=["pushed_but_master_moved", "merge_rejected"],
        )
        if fusion is Fusion.gitlab_rebase:
            rebase_url = (
                f'/projects/{mocks.mocklab.merge_request_info["project_id"]}/'
                f'merge_requests/{mocks.mocklab.merge_request_info["iid"]}/rebase'
            )

            mocks.api.add_transition(
                PUT(rebase_url),
                Ok(True),
                from_state="initial",
                to_state="pushed_but_master_moved",
                side_effect=lambda: (
                    remote_source_repo.set_ref(source_branch, first_rewritten_sha),
                    remote_target_repo.set_ref(target_branch, moved_master_sha),
                ),
            )
            mocks.api.add_transition(
                PUT(rebase_url),
                Ok(True),
                from_state="merge_rejected",
                to_state="rebase-in-progress",
                side_effect=lambda: remote_source_repo.set_ref(
                    source_branch, second_rewritten_sha
                ),
            )

        mocks.job.execute()
        assert mocks.api.state == "merged"
        assert mocks.api.notes == [
            "A merge request may have skipped the queue, trying again..."
        ]

    def test_handles_races_for_merging(self, mocks):
        rewritten_sha = mocks.mocklab.rewritten_sha
        mocks.api.add_transition(
            PUT(
                f"/projects/1234/merge_requests/"
                f'{mocks.mocklab.merge_request_info["iid"]}/merge',
                {
                    "sha": rewritten_sha,
                    "should_remove_source_branch": True,
                    "merge_when_pipeline_succeeds": True,
                    "squash": False,
                },
            ),
            Error(marge.gitlab.NotFound(404, {"message": "404 Branch Not Found"})),
            from_state="passed",
            to_state="someone_else_merged",
        )
        mocks.api.add_merge_request(
            dict(mocks.mocklab.merge_request_info, state="merged"),
            from_state="someone_else_merged",
        )
        mocks.job.execute()
        assert mocks.api.state == "someone_else_merged"
        assert mocks.api.notes == []

    @pytest.mark.parametrize("only_allow_merge_if_pipeline_succeeds", [True, False])
    def test_calculates_merge_when_pipeline_succeeds_correctly(
        self, mocks, only_allow_merge_if_pipeline_succeeds
    ):
        rewritten_sha = mocks.mocklab.rewritten_sha
        project_info = dict(TEST_PROJECT_INFO)
        project_info["only_allow_merge_if_pipeline_succeeds"] = (
            only_allow_merge_if_pipeline_succeeds
        )
        mocks.api.add_project(project_info)
        mocks.api.add_transition(
            PUT(
                f'/projects/{project_info["id"]}/merge_requests/'
                f'{mocks.mocklab.merge_request_info["iid"]}/merge',
                {
                    "sha": rewritten_sha,
                    "should_remove_source_branch": True,
                    "merge_when_pipeline_succeeds": only_allow_merge_if_pipeline_succeeds,
                    "squash": False,
                },
            ),
            Ok(True),
            to_state="merged",
        )
        mocks.job.execute()
        assert mocks.api.state == "merged"

    def test_handles_request_becoming_draft_after_push(self, mocks):
        rewritten_sha = mocks.mocklab.rewritten_sha
        mocks.api.add_transition(
            PUT(
                f"/projects/1234/merge_requests/"
                f'{mocks.mocklab.merge_request_info["iid"]}/merge',
                {
                    "sha": rewritten_sha,
                    "should_remove_source_branch": True,
                    "merge_when_pipeline_succeeds": True,
                    "squash": False,
                },
            ),
            Error(
                marge.gitlab.MethodNotAllowed(
                    405, {"message": "405 Method Not Allowed"}
                )
            ),
            from_state="passed",
            to_state="now_is_draft",
        )
        mocks.api.add_merge_request(
            dict(mocks.mocklab.merge_request_info, draft=True),
            from_state="now_is_draft",
        )
        message = (
            "The request was marked as Draft as the bot was processing it."
            + " Maybe a Draft commit was pushed?"
        )
        with mocks.mocklab.expected_failure(message):
            mocks.job.execute()
        assert mocks.api.state == "now_is_draft"
        assert mocks.api.notes == [f"This branch couldn't be merged: {message}"]

    def test_guesses_git_hook_error_on_merge_refusal(self, mocks):
        rewritten_sha = mocks.mocklab.rewritten_sha
        mocks.api.add_transition(
            PUT(
                f"/projects/1234/merge_requests/"
                f'{mocks.mocklab.merge_request_info["iid"]}/merge',
                {
                    "sha": rewritten_sha,
                    "should_remove_source_branch": True,
                    "merge_when_pipeline_succeeds": True,
                    "squash": False,
                },
            ),
            Error(
                marge.gitlab.MethodNotAllowed(
                    405, {"message": "405 Method Not Allowed"}
                )
            ),
            from_state="passed",
            to_state="rejected_by_git_hook",
        )
        mocks.api.add_merge_request(
            dict(mocks.mocklab.merge_request_info, state="reopened"),
            from_state="rejected_by_git_hook",
        )
        message = (
            "GitLab could not merge this branch. Possibly a Push Rule or a git-hook "
            "is rejecting the commits; maybe the bot's email needs to be allow-listed?"
        )
        with mocks.mocklab.expected_failure(message):
            mocks.job.execute()
        assert mocks.api.state == "rejected_by_git_hook"
        assert mocks.api.notes == [f"This branch couldn't be merged: {message}"]

    def test_assumes_unresolved_discussions_on_merge_refusal(self, mocks):
        rewritten_sha = mocks.mocklab.rewritten_sha
        mocks.api.add_transition(
            PUT(
                f"/projects/1234/merge_requests/"
                f'{mocks.mocklab.merge_request_info["iid"]}/merge',
                {
                    "sha": rewritten_sha,
                    "should_remove_source_branch": True,
                    "merge_when_pipeline_succeeds": True,
                    "squash": False,
                },
            ),
            Error(
                marge.gitlab.MethodNotAllowed(
                    405, {"message": "405 Method Not Allowed"}
                )
            ),
            from_state="passed",
            to_state="unresolved_discussions",
        )
        mocks.api.add_merge_request(
            dict(mocks.mocklab.merge_request_info), from_state="unresolved_discussions"
        )
        message = (
            "Gitlab could not merge this branch for an unknown reason. "
            "Maybe there are unresolved discussions?"
        )
        with mocks.mocklab.expected_failure(message):
            with patch.dict(
                mocks.mocklab.project_info,
                only_allow_merge_if_all_discussions_are_resolved=True,
            ):
                mocks.job.execute()
        assert mocks.api.state == "unresolved_discussions"
        assert mocks.api.notes == [f"This branch couldn't be merged: {message}"]

    def test_discovers_if_someone_closed_the_merge_request(self, mocks):
        rewritten_sha = mocks.mocklab.rewritten_sha
        mocks.api.add_transition(
            PUT(
                f"/projects/1234/merge_requests/"
                f'{mocks.mocklab.merge_request_info["iid"]}/merge',
                {
                    "sha": rewritten_sha,
                    "should_remove_source_branch": True,
                    "merge_when_pipeline_succeeds": True,
                    "squash": False,
                },
            ),
            Error(
                marge.gitlab.MethodNotAllowed(
                    405, {"message": "405 Method Not Allowed"}
                )
            ),
            from_state="passed",
            to_state="oops_someone_closed_it",
        )
        mocks.api.add_merge_request(
            dict(mocks.mocklab.merge_request_info, state="closed"),
            from_state="oops_someone_closed_it",
        )
        message = "The merge request was closed while the merge was being attempted."
        with mocks.mocklab.expected_failure(message):
            mocks.job.execute()
        assert mocks.api.state == "oops_someone_closed_it"
        assert mocks.api.notes == [f"This branch couldn't be merged: {message}"]

    def test_tells_explicitly_that_gitlab_refused_to_merge(self, mocks):
        rewritten_sha = mocks.mocklab.rewritten_sha
        mocks.api.add_transition(
            PUT(
                f"/projects/1234/merge_requests/"
                f'{mocks.mocklab.merge_request_info["iid"]}/merge',
                {
                    "sha": rewritten_sha,
                    "should_remove_source_branch": True,
                    "merge_when_pipeline_succeeds": True,
                    "squash": False,
                },
            ),
            Error(
                marge.gitlab.MethodNotAllowed(
                    405, {"message": "405 Method Not Allowed"}
                )
            ),
            from_state="passed",
            to_state="rejected_for_mysterious_reasons",
        )
        message = "Gitlab could not merge this branch for an unknown reason."
        with mocks.mocklab.expected_failure(message):
            mocks.job.execute()
        assert mocks.api.state == "rejected_for_mysterious_reasons"
        assert mocks.api.notes == [f"This branch couldn't be merged: {message}"]

    def test_wont_merge_draft_stuff(self, mocks):
        draft_merge_request = dict(mocks.mocklab.merge_request_info, draft=True)
        mocks.api.add_merge_request(draft_merge_request, from_state="initial")

        with mocks.mocklab.expected_failure(
            "It is not possible to merge MRs marked as Draft. "
            "Please undraft it before assigning to Marge next time."
        ):
            mocks.job.execute()

        assert mocks.api.state == "initial"
        assert mocks.api.notes == [
            "This branch couldn't be merged: "
            "It is not possible to merge MRs marked as Draft. "
            "Please undraft it before assigning to Marge next time."
        ]

    def test_wont_merge_branches_with_autosquash_if_rewriting(self, mocks_factory):
        mocks = mocks_factory(
            extra_mocklab_opts={"merge_request_options": {"squash_on_merge": True}}
        )

        admin_user = dict(mocks.mocklab.user_info, is_admin=True)
        mocks.api.add_user(admin_user, is_current=True)

        if mocks.job.opts.requests_commit_tagging:
            message = (
                "Merging requests marked as auto-squash is not possible due to configuration. "
                "Please disable squashing or talk with the maintainers about the commit tagging config."
            )
            with mocks.mocklab.expected_failure(message):
                mocks.job.execute()
            assert mocks.api.state == "initial"
        else:
            mocks.job.execute()
            assert mocks.api.state == "merged"

    @patch("marge.job.log", autospec=True)
    def test_waits_for_approvals(self, mock_log, mocks_factory):
        five_secs = timedelta(seconds=5)
        mocks = mocks_factory(
            extra_opts={"approval_timeout": five_secs, "reapprove": True}
        )
        mocks.job.execute()

        mock_log.info.assert_any_call("Checking if approvals have reset")
        mock_log.debug.assert_any_call(
            "Approvals haven't reset yet, sleeping for %s secs", ANY
        )
        assert mocks.api.state == "merged"

    def test_fails_if_changes_already_exist(self, mocks):
        source_project_info = (
            mocks.mocklab.forked_project_info or mocks.mocklab.project_info
        )
        source_project_url = source_project_info["ssh_url_to_repo"]
        target_project_url = mocks.mocklab.project_info["ssh_url_to_repo"]
        remote_source_repo = mocks.job.repo.mock_impl.remote_repos[source_project_url]
        remote_target_repo = mocks.job.repo.mock_impl.remote_repos[target_project_url]
        source_branch = mocks.mocklab.merge_request_info["source_branch"]
        target_branch = mocks.mocklab.merge_request_info["target_branch"]

        remote_target_repo.set_ref(
            target_branch, remote_source_repo.get_ref(source_branch)
        )
        expected_message = f"these changes already exist in branch `{target_branch}`"

        with mocks.mocklab.expected_failure(expected_message):
            mocks.job.execute()

        assert mocks.api.state == "initial"
        assert mocks.api.notes == [
            f"This branch couldn't be merged: {expected_message}"
        ]

    def test_git_timeout_message(self, mocks):
        timeout = 10

        def fail_on_fetch(*args):
            if "fetch" in args:
                raise subprocess.TimeoutExpired(cmd=args, timeout=timeout)

        mocks.job.repo.git = fail_on_fetch

        with pytest.raises(MockedEndpointNotFound, match=f"{timeout} seconds"):
            mocks.job.execute()



#!/usr/bin/env python3
"""Executable script to pass to git rebase to keep the original committer data."""

import os
import subprocess
import sys
from typing import NoReturn

STDERR = sys.stderr.buffer


def die(msg: str) -> NoReturn:
    STDERR.write(b"ERROR: ")
    STDERR.write(msg.encode())
    STDERR.write(b"\n")
    sys.exit(1)


def get_log_info(git_format: str, oid: str) -> str:
    subp = subprocess.run(
        ["git", "log", f"--format={git_format}", "-1", oid],
        capture_output=True,
        check=False,
        text=True,
    )
    if subp.returncode:
        die(subp.stderr)
    return subp.stdout.strip()


def get_committer_env(oid: str) -> dict[str, str]:
    env = {}
    env["GIT_COMMITTER_DATE"] = get_log_info("%cD", oid)
    env["GIT_COMMITTER_NAME"] = get_log_info("%cn", oid)
    env["GIT_COMMITTER_EMAIL"] = get_log_info("%ce", oid)
    return env


def main(oid: str) -> int:
    committer_env = os.environ.copy()
    committer_env.update(get_committer_env(oid))
    subp = subprocess.run(
        ["git", "commit", "--amend", "--no-edit", "--allow-empty"],
        capture_output=True,
        check=False,
        env=committer_env,
        text=True,
    )
    if subp.returncode:
        die(subp.stderr)
    return 0


if __name__ == "__main__":
    if len(sys.argv) < 2:
        die("Missing OID (sha) argument")
    sys.exit(main(sys.argv[1]))



import sys

from marge.app import main


def run() -> int:
    try:
        return main()
    except Exception as err:
        print("Exception occurred")
        if hasattr(err, "stdout"):
            # pylint: disable=no-member
            print(f"stdout was: {err.stdout}")
        if hasattr(err, "stderr"):
            # pylint: disable=no-member
            print(f"stderr was: {err.stderr}")
        raise


if __name__ == "__main__":
    sys.exit(run())



import dataclasses
import logging as log
import re
from typing import Any

from marge import gitlab
from tests import test_approvals, test_commit, test_project, test_user

GET = gitlab.GET
POST = gitlab.POST


def commit(commit_id, status):
    return {
        "id": commit_id,
        "short_id": commit_id,
        "author_name": "J. Bond",
        "author_email": "jbond@mi6.gov.uk",
        "message": "Shaken, not stirred",
        "status": status,
    }


class MockLab:  # pylint: disable=too-few-public-methods
    def __init__(
        self,
        initial_master_sha="505e",
        gitlab_url=None,
        fork=False,
        merge_request_options=None,
    ):
        self.gitlab_url = gitlab_url = gitlab_url or "http://git.example.com"
        self.api = api = Api(
            gitlab_url=gitlab_url, auth_token="no-token", initial_state="initial"
        )

        self.user_info = dict(test_user.INFO)
        self.user_id = self.user_info["id"]
        api.add_user(self.user_info, is_current=True)

        self.project_info = dict(test_project.INFO)
        api.add_project(self.project_info)

        self.commit_info = dict(test_commit.INFO)
        api.add_commit(self.project_info["id"], self.commit_info)

        self.author_id = 234234
        self.merge_request_info = {
            "id": 53,
            "iid": 54,
            "title": "a title",
            "project_id": 1234,
            "author": {"id": self.author_id},
            "assignees": [{"id": self.user_id}],
            "state": "opened",
            "detailed_merge_status": "mergeable",
            "sha": self.commit_info["id"],
            "source_project_id": 1234,
            "target_project_id": 1234,
            "source_branch": "useless_new_feature",
            "force_remove_source_branch": True,
            "target_branch": "master",
            "draft": False,
            "blocking_discussions_resolved": True,
            "web_url": "http://git.example.com/group/project/merge_request/666",
        }
        if merge_request_options is not None:
            self.merge_request_info.update(merge_request_options)

        if fork:
            self.forked_project_info = dict(
                self.project_info,
                id=4321,
                ssh_url_to_repo="ssh://some.other.project/stuff",
            )
            api.add_project(self.forked_project_info)
            self.merge_request_info.update({"iid": 55, "source_project_id": "4321"})
        else:
            self.forked_project_info = None

        api.add_merge_request(self.merge_request_info)

        self.initial_master_sha = initial_master_sha
        self.approvals_info = dict(
            test_approvals.INFO,
            id=self.merge_request_info["id"],
            iid=self.merge_request_info["iid"],
            project_id=self.merge_request_info["project_id"],
            approvals_left=0,
        )
        api.add_approvals(self.approvals_info)
        api.add_transition(
            GET(
                f"/projects/1234/repository/branches/"
                f'{self.merge_request_info["target_branch"]}'
            ),
            Ok({"commit": {"id": self.initial_master_sha}}),
        )


class Api(gitlab.Api):
    def __init__(self, gitlab_url, auth_token, initial_state):
        super().__init__(gitlab_url, auth_token)

        self._transitions = {}
        self.state = initial_state
        self.notes = []

    def call(self, command, sudo=None):
        sudo_str = sudo if sudo is not None else ""
        log.info("CALL: %s%s @ %s", f"sudo {sudo_str} ", command, self.state)
        try:
            response, next_state, side_effect = self._find(command, sudo)
        except KeyError as err:
            page = command.args.get("page")
            if page == 0:
                no_page_args = dict(
                    (k, v)
                    for k, v in command.args.items()
                    if k not in ["page", "per_page"]
                )
                try:
                    return self.call(dataclasses.replace(command, args=no_page_args))
                except MockedEndpointNotFound:
                    pass  # raise the right exception below
            elif page:  # page is not None
                try:
                    # only return an empty list if the command exists
                    self.call(command.for_page(0))
                except MockedEndpointNotFound:
                    pass  # raise the right exception below
                else:
                    return []

            raise MockedEndpointNotFound(command, sudo, self.state) from err

        if next_state:
            self.state = next_state

        if side_effect:
            side_effect()
        return response()

    def _find(self, command, sudo):
        more_specific = self._transitions.get(_key(command, sudo, self.state))
        return more_specific or self._transitions[_key(command, sudo, None)]

    def add_transition(
        self,
        command,
        response,
        sudo=None,
        from_state=None,
        to_state=None,
        side_effect=None,
    ):
        from_states = from_state if isinstance(from_state, list) else [from_state]

        for _from_state in from_states:
            show_from = "*" if _from_state is None else repr(_from_state)
            sudo_str = sudo if sudo is not None else ""
            log.info(
                "REGISTERING %s%s from %s to %s",
                f"sudo {sudo_str} ",
                command,
                show_from,
                show_from if to_state is None else repr(to_state),
            )
            self._transitions[_key(command, sudo, _from_state)] = (
                response,
                to_state,
                side_effect,
            )

    def add_resource(self, path, info, sudo=None, from_state=None, to_state=None):
        self.add_transition(
            GET(path.format(attrs(info))), Ok(info), sudo, from_state, to_state
        )

    def add_user(
        self, info, is_current=False, sudo=None, from_state=None, to_state=None
    ):
        self.add_resource("/users/{0.id}", info, sudo, from_state, to_state)
        if is_current:
            self.add_resource("/user", info, sudo, from_state, to_state)

    def add_project(self, info, sudo=None, from_state=None, to_state=None):
        self.add_resource("/projects/{0.id}", info, sudo, from_state, to_state)
        self.add_transition(
            GET(f"/projects/{attrs(info).id}/merge_requests"),
            List(r"/projects/\d+/merge_requests/\d+$", self),
            sudo,
            from_state,
            to_state,
        )

    def add_merge_request(self, info, sudo=None, from_state=None, to_state=None):
        self.add_resource(
            "/projects/{0.project_id}/merge_requests/{0.iid}",
            info,
            sudo,
            from_state,
            to_state,
        )
        self.add_transition(
            GET(
                f"/projects/{attrs(info).project_id}/merge_requests/{attrs(info).iid}",
                args={"include_rebase_in_progress": "true"},
            ),
            Ok(info),
            sudo,
            from_state,
            to_state,
        )

    def add_commit(self, project_id, info, sudo=None, from_state=None, to_state=None):
        path = f"/projects/{project_id}/repository/commits/{{0.id}}"
        self.add_resource(path, info, sudo, from_state, to_state)

    def add_approvals(self, info, sudo=None, from_state=None, to_state=None):
        path = "/projects/{0.project_id}/merge_requests/{0.iid}/approvals"
        self.add_resource(path, info, sudo, from_state, to_state)

    def add_pipelines(
        self,
        project_id,
        merge_request_iid,
        info,
        sudo=None,
        from_state=None,
        to_state=None,
    ):
        self.add_transition(
            GET(f"/projects/{project_id}/merge_requests/{merge_request_iid}/pipelines"),
            Ok([info]),
            sudo,
            from_state,
            to_state,
        )

    def expected_note(
        self, merge_request, note, sudo=None, from_state=None, to_state=None
    ):
        self.add_transition(
            POST(
                f"/projects/{attrs(merge_request).project_id}/merge_requests/"
                f"{attrs(merge_request).iid}/notes",
                args={"body": note},
            ),
            LeaveNote(note, self),
            sudo,
            from_state,
            to_state,
        )


def _key(command, sudo, state):
    return (
        dataclasses.replace(command, args=frozenset(command.args.items())),
        sudo,
        state,
    )


@dataclasses.dataclass
class Ok:
    result: dict[str, Any]

    def __call__(self):
        return self.result


@dataclasses.dataclass
class Error:
    exc: Exception

    def __call__(self):
        raise self.exc


@dataclasses.dataclass
class List:
    prefix: str
    api: Api

    def __call__(self):
        candidates = (
            command
            for command, _ in self.api._transitions  # pylint: disable=protected-access
            if isinstance(command, GET) and re.match(self.prefix, command.endpoint)
        )

        results = []
        for command in candidates:
            try:
                results.append(self.api.call(command))
            except MockedEndpointNotFound:
                pass

        return results


@dataclasses.dataclass
class LeaveNote:
    note: str
    api: Api

    def __call__(self):
        self.api.notes.append(self.note)
        return {}


class MockedEndpointNotFound(Exception):
    pass


def attrs(_dict):
    attrs_cls = dataclasses.make_dataclass("Attrs", fields=list(_dict.keys()))
    return attrs_cls(**_dict)



import enum
from typing import TYPE_CHECKING

from . import gitlab

GET = gitlab.GET


class Project(gitlab.Resource):
    @classmethod
    def fetch_by_id(cls, project_id: int, api: gitlab.Api) -> "Project":
        info = api.call(GET(f"/projects/{project_id}"))
        if TYPE_CHECKING:
            assert isinstance(info, dict)
        return cls(api, info)

    @property
    def id(self) -> int:
        result = self._info["id"]
        if TYPE_CHECKING:
            assert isinstance(result, int)
        return result

    @property
    def default_branch(self) -> str:
        result = self.info["default_branch"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def path_with_namespace(self) -> str:
        result = self.info["path_with_namespace"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def ssh_url_to_repo(self) -> str:
        result = self.info["ssh_url_to_repo"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def http_url_to_repo(self) -> str:
        result = self.info["http_url_to_repo"]
        if TYPE_CHECKING:
            assert isinstance(result, str)
        return result

    @property
    def merge_requests_enabled(self) -> bool:
        result = self.info["merge_requests_enabled"]
        if TYPE_CHECKING:
            assert isinstance(result, bool)
        return result

    @property
    def only_allow_merge_if_pipeline_succeeds(self) -> bool:
        result = self.info["only_allow_merge_if_pipeline_succeeds"]
        if TYPE_CHECKING:
            assert isinstance(result, bool)
        return result

    @property
    def only_allow_merge_if_all_discussions_are_resolved(  # pylint: disable=invalid-name
        self,
    ) -> bool:
        result = self.info["only_allow_merge_if_all_discussions_are_resolved"]
        if TYPE_CHECKING:
            assert isinstance(result, bool)
        return result

    @property
    def approvals_required(self) -> int:
        result = self.info["approvals_before_merge"]
        if TYPE_CHECKING:
            assert isinstance(result, int)
        return result

    @property
    def access_level(self) -> int:
        permissions = self.info["permissions"]
        effective_access = (
            permissions["project_access"]
            or permissions["group_access"]
            or permissions.get("marge")
        )
        assert (
            effective_access is not None
        ), "GitLab failed to provide user permissions on project"
        return AccessLevel(effective_access["access_level"])

    @property
    def archived(self) -> bool:
        result = self.info["archived"]
        if TYPE_CHECKING:
            assert isinstance(result, bool)
        return result


# pylint: disable=invalid-name
@enum.unique
class AccessLevel(enum.IntEnum):
    # See https://docs.gitlab.com/ce/api/access_requests.html
    none = 0
    minimal = 5
    guest = 10
    reporter = 20
    developer = 30
    maintainer = 40
    owner = 50
